from operator import itemgetter

import fsui
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from fsgs.plugins.pluginmanager import PluginManager
from fsui import Image
from launcher.devicemanager import DeviceManager
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.settings.settings_dialog import SettingsDialog
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior
from launcher.ui.config.configdialog import ConfigDialog
from launcher.ui.download import DownloadGameWindow
from launcher.ui.kickstartimportdialog import KickstartImportDialog
from launcher.ui.statusbar.StatusElement import StatusElement
from launcher.update_manager import UpdateManager

ERROR_LEVEL = 0
WARNING_LEVEL = 1
NOTICE_LEVEL = 2
JOYSTICK_KEYS = [
    "joystick_port_0",
    "joystick_port_1",
    "joystick_port_2",
    "joystick_port_3",
    "joystick_port_0_mode",
    "joystick_port_1_mode",
    "joystick_port_2_mode",
    "joystick_port_3_mode",
]


class WarningsElement(StatusElement):
    def __init__(self, parent):
        StatusElement.__init__(self, parent)
        self.error_icon = Image("launcher:res/16x16/error.png")
        self.warning_icon = Image("launcher:res/16x16/warning_3.png")
        self.notice_icon = Image("launcher:res/16x16/information.png")
        self.icons = [self.error_icon, self.warning_icon, self.notice_icon]
        self.coordinates = []
        self.warnings = []
        self.game_notice = ""
        self.variant_notice = ""
        self.variant_warning = ""
        self.variant_error = ""
        self.joy_emu_conflict = ""
        self.using_joy_emu = False
        self.kickstart_file = ""
        self.x_kickstart_file_sha1 = ""
        self.update_available = ""
        self.__error = ""
        self.x_missing_files = ""
        self.download_page = ""
        self.download_file = ""
        self.platform = ""
        self.amiga_model = ""
        self.amiga_model_calculated = ""
        self.chip_memory = ""
        self.chip_memory_calculated = 0
        self.outdated_plugins = []
        self.custom_config = set()
        self.custom_uae_config = set()
        self.settings_config_keys = set()

        plugin_manager = PluginManager.instance()
        for plugin in plugin_manager.plugins():
            if plugin.outdated:
                self.outdated_plugins.append(plugin.name)

        ConfigBehavior(
            self,
            [
                "x_game_notice",
                "x_variant_notice",
                "x_variant_warning",
                "x_variant_error",
                "x_joy_emu_conflict",
                "amiga_model",
                "x_kickstart_file_sha1",
                "kickstart_file",
                "download_page",
                "download_file",
                "x_missing_files",
                "__error",
                "chip_memory",
                "jit_compiler",
                "platform",
            ],
        )
        SettingsBehavior(self, ["__update_available"])

        LauncherConfig.add_listener(self)
        for key in JOYSTICK_KEYS:
            self.on_config(key, LauncherConfig.get(key))
        for key in LauncherConfig.keys():
            if LauncherConfig.is_custom_uae_option(key):
                self.on_config(key, LauncherConfig.get(key))
            elif LauncherConfig.is_custom_option(key):
                self.on_config(key, LauncherConfig.get(key))

        LauncherSettings.add_listener(self)
        for key in LauncherSettings.keys():
            if LauncherConfig.is_config_only_option(key):
                self.on_setting(key, LauncherSettings.get(key))

    def on_destroy(self):
        LauncherConfig.remove_listener(self)
        LauncherSettings.remove_listener(self)

    def on_platform_config(self, value):
        if value != self.platform:
            self.platform = value
            self.rebuild_warnings_and_refresh()

    def on_amiga_model_config(self, value):
        LauncherConfig.update_kickstart()
        if value != self.amiga_model:
            self.amiga_model = value
            self.amiga_model_calculated = value.split("/")[0]
            self.rebuild_warnings_and_refresh()

    def on_chip_memory_config(self, value):
        if value != self.chip_memory:
            self.chip_memory = value
            try:
                self.chip_memory_calculated = int(value or "0")
            except Exception:
                self.chip_memory_calculated = -1
            self.rebuild_warnings_and_refresh()

    def on___error_config(self, value):
        if value != self.__error:
            self.__error = value
            self.rebuild_warnings_and_refresh()

    def on_x_missing_files_config(self, value):
        if value != self.x_missing_files:
            self.x_missing_files = value
            self.rebuild_warnings_and_refresh()

    def on_download_page_config(self, value):
        if value != self.download_page:
            self.download_page = value
            self.rebuild_warnings_and_refresh()

    def on_download_file_config(self, value):
        if value != self.download_file:
            self.download_file = value
            self.rebuild_warnings_and_refresh()

    def on_kickstart_file_config(self, value):
        if value != self.kickstart_file:
            self.kickstart_file = value
            self.rebuild_warnings_and_refresh()

    def on_x_kickstart_file_sha1_config(self, value):
        if value != self.x_kickstart_file_sha1:
            self.x_kickstart_file_sha1 = value
            self.rebuild_warnings_and_refresh()

    def on_x_game_notice_config(self, value):
        if value != self.game_notice:
            self.game_notice = value
            self.rebuild_warnings_and_refresh()

    def on_x_variant_notice_config(self, value):
        if value != self.variant_notice:
            self.variant_notice = value
            self.rebuild_warnings_and_refresh()

    def on_x_variant_warning_config(self, value):
        if value != self.variant_warning:
            self.variant_warning = value
            self.rebuild_warnings_and_refresh()

    def on_x_variant_error_config(self, value):
        if value != self.variant_error:
            self.variant_error = value
            self.rebuild_warnings_and_refresh()

    def on_x_joy_emu_conflict_config(self, value):
        print("\n\n\nGOT x_joy_emu_conflict\n\n\n")
        if value != self.joy_emu_conflict:
            self.joy_emu_conflict = value
            self.rebuild_warnings_and_refresh()

    def on_jit_compiler_config(self, _):
        self.rebuild_warnings_and_refresh()

    def on_config(self, key, value):
        if key in JOYSTICK_KEYS:
            prev_value = self.using_joy_emu
            devices = DeviceManager.get_devices_for_ports(LauncherConfig)
            for device in devices:
                if device.id == "keyboard":
                    self.using_joy_emu = True
                    break
            else:
                self.using_joy_emu = False
            if prev_value != self.using_joy_emu:
                self.rebuild_warnings_and_refresh()
        elif key.startswith("__"):
            pass
        # elif LauncherConfig.is_implicit_option(key):
        #     pass
        elif LauncherConfig.is_custom_uae_option(key):
            changed = False
            if value:
                if key not in self.custom_uae_config:
                    self.custom_uae_config.add(key)
                    changed = True
            else:
                if key in self.custom_uae_config:
                    self.custom_uae_config.remove(key)
                    changed = True
            if changed:
                self.rebuild_warnings_and_refresh()
        elif LauncherConfig.is_custom_option(key):
            changed = False
            if value:
                if key not in self.custom_config:
                    self.custom_config.add(key)
                    changed = True
            else:
                if key in self.custom_config:
                    self.custom_config.remove(key)
                    changed = True
            if changed:
                self.rebuild_warnings_and_refresh()

    def on_setting(self, key, value):
        if LauncherConfig.is_config_only_option(key):
            changed = False
            if value:
                if key not in self.settings_config_keys:
                    self.settings_config_keys.add(key)
                    changed = True
            else:
                if key in self.settings_config_keys:
                    self.settings_config_keys.remove(key)
                    changed = True
            if changed:
                self.rebuild_warnings_and_refresh()

    def on___update_available_setting(self, value):
        if value != self.update_available:
            self.update_available = value
            self.rebuild_warnings_and_refresh()

    def paint_element(self, dc):
        self.coordinates.clear()
        dc.set_font(dc.get_font())
        x = 6
        w, h = self.size()
        for level, warning, handler in self.warnings:
            icon = self.icons[level]
            start = x
            dc.draw_image(icon, x, 6)
            x += 16 + 6
            tw, th = dc.measure_text(warning)
            dc.draw_text(warning, x, round((h - th) / 2))
            x += tw + 6 + 16
            stop = x
            self.coordinates.append((start, stop, handler))

    def rebuild_warnings(self):
        self.warnings = []

        if self.using_joy_emu and self.joy_emu_conflict:
            self.warnings.append((WARNING_LEVEL, self.joy_emu_conflict, ""))

        self.add_option_warnings()
        self.add_game_warnings()

        if self.update_available:
            text = gettext("Update available: {version}").format(
                version=self.update_available
            )
            self.warnings.append((NOTICE_LEVEL, text, "on_update"))

        if self.outdated_plugins:
            text = gettext(
                "Outdated plugins: {0}".format(
                    ", ".join(self.outdated_plugins)
                )
            )
            self.warnings.append((ERROR_LEVEL, text, "on_outdated_plugins"))

        if (
            self.platform in ["", "amiga", "cdtv", "cd32"]
            and self.x_kickstart_file_sha1 == Amiga.INTERNAL_ROM_SHA1
            and self.kickstart_file != "internal"
        ):
            # text = gettext("Compatibility Issue")
            # self.warnings.append((ERROR_LEVEL, text, "on_kickstart_warning"))
            text = gettext("Using Kickstart ROM replacement")
            self.warnings.append((WARNING_LEVEL, text, "on_kickstart_warning"))
            text = gettext("Click to import Kickstart ROMs")
            self.warnings.append((NOTICE_LEVEL, text, "on_import_kickstarts"))

        if self.__error:
            self.warnings.append((ERROR_LEVEL, self.__error, ""))

        self.add_config_warnings()
        self.warnings.sort(key=itemgetter(0))

    def add_option_warnings(self):
        if len(self.settings_config_keys):
            if len(self.settings_config_keys) == 1:
                text = gettext(
                    "Config in settings: {name}".format(
                        name=list(self.settings_config_keys)[0]
                    )
                )
            else:
                text = gettext("Config options in settings")
            self.warnings.append((ERROR_LEVEL, text, "on_advanced_settings"))
        if len(self.custom_uae_config):
            if len(self.custom_uae_config) == 1:
                text = gettext(
                    "Custom option: {name}".format(
                        name=list(self.custom_uae_config)[0]
                    )
                )
            else:
                text = gettext("Custom uae_ Options")
            self.warnings.append((WARNING_LEVEL, text, "on_uae_config"))
        if len(self.custom_config):
            if len(self.custom_config) == 1:
                text = gettext(
                    "Custom option: {name}".format(
                        name=list(self.custom_config)[0]
                    )
                )
            else:
                text = gettext("Custom options")
            self.warnings.append((NOTICE_LEVEL, text, "on_uae_config"))

    def add_game_warnings(self):
        if is_warning(self.x_missing_files):
            if self.download_file:
                text = gettext("Auto-download")
                self.warnings.append((NOTICE_LEVEL, text, ""))
            elif self.download_page:
                text = gettext("Download game")
                self.warnings.append((WARNING_LEVEL, text, "on_download_page"))
            else:
                text = gettext("Missing game files")
                self.warnings.append((ERROR_LEVEL, text, ""))

        for name in ["variant_notice", "game_notice"]:
            value = getattr(self, name)
            if not value:
                continue
            if value.startswith("WARNING: "):
                level = WARNING_LEVEL
                message = value[9:]
            else:
                level = NOTICE_LEVEL
                message = value
            self.warnings.append((level, message, ""))
        if self.variant_warning:
            self.warnings.append((WARNING_LEVEL, self.variant_warning, ""))
        if self.variant_error:
            self.warnings.append((ERROR_LEVEL, self.variant_error, ""))

    def add_config_warnings(self):
        # FIXME: move such warnings to config model code instead
        if (
            self.chip_memory_calculated
            and self.chip_memory_calculated < 2048
            and self.amiga_model_calculated in ["A1200", "A4000"]
        ):
            text = gettext("{amiga_model} with < 2 MB chip memory").format(
                amiga_model=self.amiga_model
            )
            self.warnings.append((WARNING_LEVEL, text, ""))
        if LauncherConfig.get("amiga_model") == "A4000/OS4":
            if LauncherConfig.get("jit_compiler") == "1":
                text = gettext(
                    "JIT compiler with a PPC-only OS is not recommended"
                )
                self.warnings.append((WARNING_LEVEL, text, ""))

    def rebuild_warnings_and_refresh(self):
        self.rebuild_warnings()
        self.refresh()

    def on_left_down(self):
        from fsui.qt import QCursor

        # noinspection PyArgumentList
        # FIXME: Remove direct Qt dependency
        p = self._widget.mapFromGlobal(QCursor.pos())
        for start, stop, handler in self.coordinates:
            if start <= p.x() < stop and handler:
                # print(start, stop, handler)
                getattr(self, handler)()

    def on_update(self):
        UpdateManager.start_update(self.update_available)

    def on_outdated_plugins(self):
        dialog = SettingsDialog.open(self.get_window())
        # FIXME: Move code to settings dialog, so we can open the page with
        # a method/constant or method instead of a name.
        dialog.set_page_by_title(gettext("Plugins"))

    # noinspection PyMethodMayBeStatic
    def on_kickstart_warning(self):
        text = (
            "The Kickstart ROM for the chosen Amiga model was not found "
            "on your system.\n\n"
            "A replacement Kickstart ROM from the AROS project is used "
            "instead. Compatibility will be lower than if you use an "
            "original Kickstart ROM.\n\n"
            "You can use the file database scan function or the import "
            "wizards if you own the Kickstart ROM.\n\n"
            "Original Kickstart ROMs can be purchased online as part "
            "of Cloanto's Amiga Forever package, or you can extract "
            "a Kickstart from a real Amiga.\n\n"
            "If you want to use the replacement Kickstart, you can "
            "ignore this warning, or explicitly change the Kickstart "
            'to "Internal" to dismiss this warning.'
        )
        fsui.show_warning(text, gettext("Using Kickstart ROM Replacement"))

    def on_import_kickstarts(self):
        KickstartImportDialog(self.get_window()).show()

    def on_download_page(self):
        DownloadGameWindow(self.get_window(), fsgs).show()

    def on_uae_config(self):
        print("Custom config keys:", self.custom_config)
        print("Custom uae config keys:", self.custom_uae_config)
        ConfigDialog.run(self.get_window())

    def on_advanced_settings(self):
        dialog = SettingsDialog.open(self.get_window())
        # FIXME: Move code to settings dialog, so we can open the page with
        # a method/constant or method instead of a name.
        print("Config keys in settings:", self.settings_config_keys)
        dialog.set_page_by_title(gettext("Advanced Settings"))


def is_warning(w):
    if w is None:
        return False
    if isinstance(w, str):
        return bool(w)
    if isinstance(w, bool):
        return w
    return w[0] or w[1]



import fsui
from launcher.ui.floppiesgroup import FloppiesGroup
from launcher.ui.MediaListGroup import MediaListGroup
from launcher.ui.skin import Skin


class FloppiesPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()
        self.floppies_group = FloppiesGroup(self, 4)
        self.layout.add(self.floppies_group, fill=True)
        self.media_list_group = MediaListGroup(self, False)
        self.layout.add(self.media_list_group, expand=True, fill=True)



import logging
import os
import sys

import subprocess

import workspace.path
from fsgs.plugins.pluginmanager import PluginManager

logger = logging.getLogger("SYS")


def program_path():
    return [
        "SYS:C",
        "SYS:Utilities",
        "SYS:Rexxc",
        "SYS:System",
        "SYS:S",
        "SYS:Prefs",
        "SYS:WSStartup",
        "SYS:Tools",
        "SYS:Tools/Commodities",
    ]


def rexx_path():
    return ["SYS:S"]


def find_program(program):
    program = program.lower()
    for dir in program_path():
        try:
            for item in workspace.path.listdir(dir):
                if item.lower() == program:
                    p = workspace.path.join(dir, item, item + ".py")
                    if workspace.path.exists(p):
                        return workspace.path.join(dir, item)
        except FileNotFoundError:
            continue
    raise FileNotFoundError("Could not find program " + repr(program))


def workspace_exec(args):
    logger.debug("shell_exec %s", repr(args))
    program_name = args[0]
    program = find_program(program_name)
    logger.debug("%s", repr(program))

    module = load_python_program_module(program)

    if hasattr(module, "workspace_exec"):
        # print("running module.workspace_exec", module)
        return module.workspace_exec(args)

    # FIXME: other exception
    raise FileNotFoundError("Cannot exec")
    # if hasattr(module, "shell_open"):
    #     module.shell_open(args)
    # else:
    #     module.open()


def load_python_program_module(path):
    logger.debug("load_python_program_module %s", repr(path))
    # FIXME: Not nice, requires globally unique module names for apps.
    import importlib

    host_path = workspace.path.host(
        workspace.path.join(path, workspace.path.basename(path) + ".py")
    )
    host_dir = os.path.dirname(host_path)
    if host_dir not in sys.path:
        sys.path.insert(0, host_dir)
    module = importlib.import_module(
        os.path.splitext(os.path.basename(host_path))[0]
    )
    logger.debug("%s", repr(module))
    return module


def python_exec(script, argv):
    env = os.environ.copy()
    env["PYTHONPATH"] = ":".join(sys.path)
    p = subprocess.Popen(
        [sys.executable, script] + argv[1:],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        env=env,
        stderr=subprocess.STDOUT,
        bufsize=0,
    )
    return p


def expansion_exec(exe_name, argv):
    executable = PluginManager.instance().find_executable(exe_name)
    p = executable.popen(
        argv[1:],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
    )
    return p


def host_exec(exe, argv):
    print([exe] + argv[1:])
    p = subprocess.Popen(
        [exe] + argv[1:],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
    )
    return p



import sys

from fsbc import settings
from fsgs.option import Option


class ArcadeSettings:
    def __init__(self):
        # FIXME: Add fs game context as parameter and check settings through
        # that instead.
        pass

    def search(self):
        if "--disable-search" in sys.argv:
            return False
        return settings.get(Option.ARCADE_SEARCH) != "0"

    def shutdown_command(self):
        return settings.get(Option.ARCADE_SHUTDOWN)


class GameCenterSettings:
    @classmethod
    def get_shutdown_command(cls):
        return ArcadeSettings().shutdown_command()



import hashlib
import os

from fsgs.amiga.rommanager import ROMManager
from fsgs.archive import Archive

ZERO_SHA1 = "da39a3ee5e6b4b0d3255bfef95601890afd80709"


class ChecksumTool:
    def __init__(self, parent=None):
        self.parent = parent
        # fsui.Window.__init__(self, parent, "Checksumming")
        # self.layout = fsui.VerticalLayout()
        # label = fsui.HeadingLabel(self, "Checksumming file...")
        # self.layout.add(label, fill=True)
        # self.layout.add_spacer(6)
        # #self.center_on_parent()

    def checksum(self, path):
        print("[CHECKSUM]", repr(path))
        archive = Archive(path)
        if os.path.exists(path):
            size = os.path.getsize(path)
            if size == 0:
                # Either a real 0-byte file or a device node on a BSD
                # system (could be large). To reliably get the size we could
                # use ioctl, but we simply return the checksum for a 0-byte
                # file in either case.
                return ZERO_SHA1
        s = hashlib.sha1()
        f = archive.open(path)
        while True:
            data = f.read(65536)
            if not data:
                break
            s.update(data)
        return s.hexdigest()

    def checksum_rom(self, path):
        print("[CHECKSUM] ROM:", repr(path))
        archive = Archive(path)
        return ROMManager.decrypt_archive_rom(archive, path)["sha1"]



import logging
import os
import platform
import subprocess
import traceback
from configparser import ConfigParser, NoSectionError
from operator import attrgetter

import fsboot
from fsbc.system import System, linux, windows
from fsbc.util import Version
from fsgs.FSGSDirectories import FSGSDirectories

X86_MACHINES = ["x86", "i386", "i486", "i586", "i686"]
X86_64_MACHINES = ["x86_64", "x86-64", "amd64"]
X86_ANY_MACHINES = X86_MACHINES + X86_64_MACHINES

logger = logging.getLogger("PLUGINS")

known_plugin_versions = {
    "CAPSImg": "5.1fs3",
    "Cheats": "1.0.0",
    "DOSBox-FS": "0.74.4006fs7",
    "Fuse-FS": "1.3.3fs5",
    "GenesisPlusGX-LR": "1.7.4git",
    "Hatari-FS": "2.0.0fs1",
    "MAME-FS": "0.189fs1",
    "Mednafen-FS": "1.22.2fs0",
    "Mupen64Plus-LR": "2.5fs0git",
    "Nestopia-LR": "1.49fs0wip",
    "QEMU-UAE": "3.8.2qemu2.2.0",
    "Regina-FS": "3.9.1fs0",
    "RetroArch-FS": "1.6.7fs1",
    "UADE-FS": "2.13fs1",
    "Vice-FS": "3.3fs0",
}


class BasePlugin:
    def __init__(self, path, name, version="0.0.0"):
        self.path = path
        self.name = name
        self.version = version
        self.provides = {}
        # outdated being None implies that it is unknown
        self.outdated = None
        known_version = known_plugin_versions.get(self.name, None)
        if known_version:
            try:
                self.outdated = Version(self.version) < Version(known_version)
            except ValueError:
                pass

    def add_provide(self, key, value):
        logger.debug("%s -> %s", key, value)
        self.provides[key] = value

    def data_file_path(self, name):
        return os.path.join(self.path, "Data", name)


class Plugin(BasePlugin):
    def __init__(self, path, name, version, cp):
        super().__init__(path, name, version)
        self.load_provides(cp)

    def load_provides(self, cp):
        logger.debug("loading provides for %s %s", self.path, self.platform())
        try:
            for key, value in cp.items(self.platform()):
                self._provides[key] = value
        except NoSectionError:
            pass

    def __str__(self):
        return "<Plugin {0}".format(self.path)

    @staticmethod
    def os_name(pretty=False):
        if windows:
            if pretty:
                return "Windows"
            return "windows"
        elif System.macos:
            if pretty:
                return "macOS"
            return "macos"
        elif linux:
            # if os.environ.get("STEAM_RUNTIME", ""):
            #     if pretty:
            #         return "SteamOS"
            #     return "steamos"
            # else:
            if pretty:
                return "Linux"
            return "linux"
        else:
            if pretty:
                return "Unknown"
            return "unknown"

    @staticmethod
    def arch_name(pretty=False):
        if platform.machine().lower() in X86_ANY_MACHINES:
            if platform.architecture()[0] == "64bit":
                return "x86-64"
            else:
                return "x86"
        else:
            if pretty:
                return "Unknown"
            return "unknown"

    @classmethod
    def platform(cls):
        return "{}_{}".format(cls.os_name(), cls.arch_name())


class Expansion(BasePlugin):
    def __init__(self, path, arch_path, version=None):
        name = os.path.basename(path)
        version_path = os.path.join(arch_path, "Version.txt")
        if os.path.exists(version_path):
            with open(version_path, "r", encoding="UTF-8") as f:
                version = f.read().strip()
        else:
            # version_path = os.path.join(path, "Data", "Version.txt")
            version_path = os.path.join(path, "Version.txt")
            with open(version_path, "r", encoding="UTF-8") as f:
                version = f.read().strip()
        super().__init__(path, name, version)
        self._arch_path = arch_path

    def load_provides(self):
        logger.debug("Loading provides for %s", repr(self.path))
        if os.path.exists(self._arch_path):
            self.load_arch_provides()
        self.load_sha1_provides()

    def load_sha1_provides(self):
        data_dir = os.path.join(self.path, "Data")
        sha1sums_file = os.path.join(data_dir, "SHA1SUMS")
        if os.path.exists(sha1sums_file):
            with open(sha1sums_file, "r", encoding="UTF-8") as f:
                for line in f.readlines():
                    line = line.strip()
                    if not line:
                        continue
                    sha1, relative_path = line.split(" ", 1)
                    relative_path = relative_path.lstrip("*")
                    path = os.path.join(data_dir, relative_path)
                    self.add_provide("sha1:" + sha1, path)

    def load_arch_provides(self):
        for item in os.listdir(self._arch_path):
            path = os.path.join(self._arch_path, item)
            if windows:
                if item.endswith(".exe"):
                    self.add_provide("executable:" + item[:-4].lower(), path)
                elif path.endswith(".dll"):
                    self.add_provide("library:" + item.lower()[:-4], path)
            else:
                if path.endswith(".so"):
                    self.add_provide("library:" + item.lower()[:-3], path)
                elif os.access(path, os.X_OK):
                    self.add_provide("executable:" + item.lower(), path)
            if System.macos:
                if path.endswith(".app"):
                    app_path = os.path.join(path, "Contents", "MacOS")
                    if os.path.exists(app_path):
                        for app_item in os.listdir(app_path):
                            p = os.path.join(app_path, app_item)
                            if os.access(p, os.X_OK):
                                self.add_provide(
                                    "executable:" + app_item.lower(), p
                                )

    def executable(self, name):
        path = self.provides["executable:" + name]
        return PluginExecutable(self, path)

    def library_path(self, name):
        path = self.provides["library:" + name]
        return path

    def sha1_path(self, sha1):
        path = self.provides["sha1:" + sha1]
        return path

    def __str__(self):
        return "<Expansion {0}".format(self.path)


class PluginResource:
    def __init__(self, plugin, name):
        self.plugin = plugin
        self.name = name

    @property
    def path(self):
        p = os.path.join(self.plugin.path, self.plugin.provides[self.name])
        return p


class Executable:
    def __init__(self, path):
        self.path = path

    def popen(self, args, env=None, **kwargs):
        logger.info("[EXECUTE] %s %s", self.path, repr(args))
        # logger.debug("PluginExecutable.popen %s %s %s",
        #              repr(args), repr(env), repr(kwargs))
        args = [self.path] + args
        if os.environ.get("FSGS_STRACE", "") == "1":
            args.insert(0, "strace")
        if env:
            env = env.copy()
        else:
            env = os.environ.copy()

        if os.environ.get("FS_PLUGINS_LD_LIBRARY_PATH") == "1":
            env["LD_LIBRARY_PATH"] = os.path.dirname(self.path)

        return subprocess.Popen(args, env=env, **kwargs)


class PluginExecutable(Executable):
    def __init__(self, plugin, path):
        super().__init__(path)
        self.plugin = plugin


class PluginManager:
    @classmethod
    def instance(cls):
        if not hasattr(cls, "_instance") or cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def plugin_path(cls):
        result = []
        plugins_dir = FSGSDirectories.get_plugins_dir()
        result.append(plugins_dir)

        # Plugins dir location has changed, add several old and new paths here
        # to find plugins in both places (FS-UAE and OpenRetro style).
        plugins_dir = os.path.join(FSGSDirectories.get_base_dir(), "Plugins")
        if plugins_dir not in result:
            result.append(plugins_dir)
        plugins_dir = os.path.join(FSGSDirectories.get_data_dir(), "Plugins")
        if plugins_dir not in result:
            result.append(plugins_dir)

        # if plugins_dir and os.path.isdir(plugins_dir):
        #     result.append(plugins_dir)
        expansion_dir = os.path.join(
            FSGSDirectories.get_base_dir(), "Workspace", "Expansion"
        )
        if expansion_dir and os.path.isdir(expansion_dir):
            result.append(expansion_dir)

        if System.macos:
            system_plugins_dir = os.path.normpath(
                os.path.join(
                    fsboot.executable_dir(),
                    "..",
                    "..",
                    "..",
                    "..",
                    "..",
                    "..",
                    "Plugins",
                )
            )
            result.append(system_plugins_dir)
        else:
            system_plugins_dir = os.path.normpath(
                os.path.join(
                    fsboot.executable_dir(), "..", "..", "..", "Plugins"
                )
            )
            result.append(system_plugins_dir)
        # if os.path.isdir(system_plugins_dir):
        #     result.append(system_plugins_dir)

        return result

    def __init__(self):
        self._provides = {}
        self._plugins = []
        self._plugins_map = {}
        self.load_plugins()
        for plugin in self._plugins:
            for key, value in plugin.provides.items():
                self._provides[key] = plugin

    def plugin(self, name):
        return self._plugins_map[name]

    def plugins(self):
        return self._plugins.copy()

    def provides(self):
        return self._provides.copy()

    def load_plugins(self):
        plugin_path = self.plugin_path()
        logger.info("Executable dir: %s", fsboot.executable_dir())
        logger.info("Path: %s", plugin_path)
        logger.info("Machine: %s", platform.machine().lower())
        logger.info("Architecture: %s", platform.architecture()[0])
        logger.info(
            "Plugin OS/arch: %s/%s",
            Plugin.os_name(True),
            Plugin.arch_name(True),
        )

        # Reversing order so that later loaded plugins (earlier on path)
        # takes precedence.
        # for dir_path in reversed(plugin_path):
        # I guess we should sort the list by version number and only load the
        # newest plugin per name.
        for dir_path in plugin_path:
            if not os.path.isdir(dir_path):
                continue
            # logger.debug(dir_path)
            for name in os.listdir(dir_path):
                plugin_dir = os.path.join(dir_path, name)
                logger.debug(plugin_dir)
                if not os.path.isdir(plugin_dir):
                    continue
                try:
                    plugin = self.load_plugin(plugin_dir)
                except Exception:
                    logger.debug("Could not load %s", plugin_dir)
                    traceback.print_exc()
                    continue
                if plugin is None:
                    logger.debug("No plugin in %s", plugin_dir)
                    continue
                logger.debug("Found plugin in %s", plugin_dir)
                self._plugins.append(plugin)
                self._plugins_map[plugin.name] = plugin
            self._plugins.sort(key=attrgetter("name"))

    def load_plugin(self, plugin_dir):
        return self.load_expansion(plugin_dir)

        # logger.info("Load plugin %s", plugin_dir)
        # plugin_ini = os.path.join(plugin_dir, "plugin.ini")
        # name = os.path.basename(plugin_dir).split("_")[0]
        # if not os.path.exists(plugin_ini):
        #     return self.load_expansion(plugin_dir)
        # cp = ConfigParser()
        # with open(plugin_ini, "r", encoding="UTF-8") as f:
        #     cp.read_file(f)
        # version = cp.get("plugin", "version")
        # plugin = Plugin(os.path.join(plugin_dir, version), name, version, cp)
        # return plugin

    def load_expansion(self, path):
        logger.info("Load expansion %s", path)
        arch_path = os.path.join(
            path, Plugin.os_name(pretty=True), Plugin.arch_name(pretty=True)
        )
        version_path = os.path.join(arch_path, "Version.txt")
        if not os.path.exists(version_path):
            # version_path = os.path.join(path, "Data", "Version.txt")
            version_path = os.path.join(path, "Version.txt")

        version = None
        if not os.path.exists(version_path):
            plugin_ini = os.path.join(path, "Plugin.ini")
            if os.path.exists(plugin_ini):
                cp = ConfigParser()
                with open(plugin_ini, "r", encoding="UTF-8") as f:
                    cp.read_file(f)
                version = cp.get("plugin", "version")
            if not version:
                plugin_ini = os.path.join(path, "plugin.ini")
                if os.path.exists(plugin_ini):
                    cp = ConfigParser()
                    with open(plugin_ini, "r", encoding="UTF-8") as f:
                        cp.read_file(f)
                    version = cp.get("plugin", "version")
            if not version:
                return None

        expansion = Expansion(path, arch_path, version=version)
        expansion.load_provides()
        return expansion

    def find_resource(self, name):
        plugin = self._provides[name]
        return PluginResource(plugin, name)

    def find_executable(self, name):
        logger.debug("PluginManager.find_executable %s", repr(name))
        try:
            plugin = self.provides()["executable:" + name]
        except KeyError:
            # Did not find executable in plugin, try to find executable
            # bundled with the program.
            if windows:
                exe_name = name + ".exe"
            else:
                exe_name = name
            path = os.path.join(fsboot.executable_dir(), exe_name)
            logger.debug("Checking %s", path)
            if os.path.exists(path):
                logger.debug("Found non-plugin executable %s", path)
                return Executable(path)
            if fsboot.development():
                if name == "x64sc-fs":
                    logger.debug("Lookup hack for vice-fs/x64sc-fs")
                    name = "vice-fs"
                path = os.path.join(
                    fsboot.executable_dir(), "..", name, exe_name
                )
                logger.debug("Checking %s", path)
                if os.path.exists(path):
                    logger.debug("Found non-plugin executable %s", path)
                    return Executable(path)
            return None
        return plugin.executable(name)

    def find_library_path(self, name):
        logger.debug("PluginManager.find_library_path %s", repr(name))
        try:
            plugin = self.provides()["library:" + name]
        except KeyError:
            # Did not find module in plugin, try to find module
            # bundled with the program.
            if windows:
                module_name = name + ".dll"
            else:
                module_name = name + ".so"
            path = os.path.join(fsboot.executable_dir(), module_name)
            if os.path.exists(path):
                logger.debug("Found non-plugin module %s", path)
                return path
            return None
        return plugin.library_path(name)

    def find_file_by_sha1(self, sha1):
        try:
            plugin = self.provides()["sha1:" + sha1]
        except KeyError:
            path = None
        else:
            path = plugin.sha1_path(sha1)
        logger.debug("PluginManager.find_file_by_sha1 %s -> %s", sha1, path)
        return path



# DataDir.py - return location of my data directory

from __future__ import print_function

import os.path

def get_data_dir():
  # retrieve vamos home and data dir
  home_dir = os.path.dirname(__file__)
  data_dir = os.path.join(home_dir, "..", "data")
  data_dir = os.path.abspath(data_dir)
  if os.path.isdir(data_dir):
    return data_dir
  else:
    return None

def get_data_sub_dir(sub_name):
  data_dir = get_data_dir()
  if data_dir is None:
    raise IOError("Data dir missing: " + data_dir)
  sub_dir = os.path.join(data_dir, sub_name)
  if os.path.isdir(sub_dir):
    return sub_dir
  else:
    return None

def ensure_data_sub_dir(sub_name):
  sub_dir = get_data_sub_dir(sub_name)
  if sub_dir is None:
    raise IOError("Data sub dir missing: " + sub_dir)
  else:
    return sub_dir


# ----- mini test -----
if __name__ == '__main__':
  print("data_dir:", get_data_dir())
  print("sub_dir:", get_data_sub_dir("fd"))



from fsui.qt import QComboBox
from .widget_mixin import WidgetMixin


class ComboBox(QComboBox, WidgetMixin):
    def __init__(self, parent, items=[], read_only=False):
        QComboBox.__init__(self, parent.get_container())
        # self = QComboBox(parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        # self.setEditable(not read_only)
        print(
            "FIXME: ComboBox not respecting read_only"
            "(because of missing lineEdit then"
        )
        self.setEditable(True)
        self.lineEdit().installEventFilter(self.get_window())

        self.set_items(items)

        if len(items) > 0:
            self.set_index(0)
        self.currentIndexChanged.connect(self.__current_index_changed)

    def __current_index_changed(self):
        self.on_changed()

    def get_min_width(self):
        # We don't want min width to be decided by the contents of the
        # line edit
        return 50

    def set_items(self, items):
        self.clear()
        for i, item in enumerate(items):
            self.insertItem(i, item)

    def set_item_text(self, item, text):
        self.setItemText(item, text)

    def get_index(self):
        # return self.GetSelection()
        return self.currentIndex()

    def set_index(self, index):
        self.setCurrentIndex(index)

    def get_text(self):
        return self.lineEdit().text()

    def set_text(self, text):
        self.lineEdit().setText(text)

    def on_changed(self):
        pass



from __future__ import absolute_import
from __future__ import print_function

import time

from amitools.fs.block.Block import Block
from amitools.fs.block.UserDirBlock import UserDirBlock
from amitools.fs.block.RootBlock import RootBlock
from amitools.fs.block.FileHeaderBlock import FileHeaderBlock
from amitools.fs.block.FileListBlock import FileListBlock
from amitools.fs.block.FileDataBlock import FileDataBlock
from amitools.fs.block.BitmapBlock import BitmapBlock
from amitools.fs.block.BitmapExtBlock import BitmapExtBlock
from amitools.fs.block.CommentBlock import CommentBlock
from amitools.fs.FSString import FSString
import amitools.fs.DosType as DosType

from amitools.fs.validate.Log import Log

class BlockInfo:
  """Store essential info of a block"""
  def __init__(self, blk_num):
    self.blk_num = blk_num
    self.blk_status = BlockScan.BS_UNKNOWN
    self.blk_type = BlockScan.BT_UNKNOWN
    self.used = False
    self.own_key = None
  
  def __str__(self):
    return str(self.__dict__)

class BlockScan:
  """Scan a full volume and classify the blocks"""
  
  # block status
  BS_UNKNOWN = 0 # undecided or unchecked
  BS_READ_ERROR = 1 # error reading block
  BS_INVALID = 2 # not a detected AmigaDOS block
  BS_VALID = 3 # is a AmigaDOS block structure but type was not detected
  BS_TYPE = 4 # detected block type
  NUM_BS = 5
  
  # block type
  BT_UNKNOWN = 0
  BT_ROOT = 1
  BT_DIR = 2
  BT_FILE_HDR = 3
  BT_FILE_LIST = 4
  BT_FILE_DATA = 5
  BT_BITMAP = 6
  BT_BITMAP_EXT = 7
  BT_COMMENT = 8
  NUM_BT = 9
  
  def __init__(self, blkdev, log, dos_type):
    self.blkdev = blkdev
    self.log = log
    self.dos_type = dos_type

    self.map_status = None
    self.map_type = None
    self.block_map = [None] * self.blkdev.num_blocks
    self.map_status = [[] for i in range(self.NUM_BS)]
    self.map_type = [[] for i in range(self.NUM_BT)]

  def scan_all(self, progress=lambda x : x):
    """Scan all blocks of the given block device
       Return True if there is a chance that a file system will be found there
    """
    # range to scan
    begin_blk = self.blkdev.reserved
    num_blk = self.blkdev.num_blocks - self.blkdev.reserved
    self.log.msg(Log.DEBUG,"block: checking range: +%d num=%d" % (begin_blk, num_blk))

    # scan all blocks
    for n in range(num_blk):
      blk_num = n + begin_blk
      
      # read/get block
      bi = self.get_block(blk_num)
      
      # own key ok?
      if bi != None:
        if bi.blk_status == self.BS_TYPE:
          if bi.own_key != None and bi.own_key != blk_num:
            self.log.msg(Log.ERROR, "Own key is invalid: %d type: %d" % (bi.own_key, bi.blk_type), blk_num)
        
      # first summary after block scan
      num_error_blocks = len(self.map_status[self.BS_READ_ERROR])
      if num_error_blocks > 0:
        self.log.msg(Log.ERROR, "%d unreadable error blocks found" % num_error_blocks)
      num_valid_blocks = len(self.map_status[self.BS_VALID])
      if num_valid_blocks > 0:
        self.log.msg(Log.INFO, "%d valid but unknown blocks found" % num_valid_blocks)
      num_invalid_blocks = len(self.map_status[self.BS_INVALID])
      if num_invalid_blocks > 0:
        self.log.msg(Log.INFO, "%d invalid blocks found" % num_invalid_blocks)      
      
  def read_block(self, blk_num, is_bm=False, is_bm_ext=False):
    """read block from device, decode it, and return block info instance"""
    try:
      # read block from device
      if is_bm:
        blk = BitmapBlock(self.blkdev, blk_num)
      elif is_bm_ext:
        blk = BitmapExtBlock(self.blkdev, blk_num)
      else:
        blk = Block(self.blkdev, blk_num)
      blk.read()
      data = blk.data
      # create block info
      bi = BlockInfo(blk_num)
      # --- classify block ---
      if blk.valid:
        # block is valid AmigaDOS
        bi.blk_status = self.BS_VALID
        # --- bitmap block ---
        if is_bm:
          bi.blk_type = self.BT_BITMAP
          bi.blk_status = self.BS_TYPE
          bi.bitmap = blk.get_bitmap_data()
        # --- bitmap ext block ---
        elif is_bm_ext:
          bi.blk_type = self.BT_BITMAP_EXT
          bi.blk_status = self.BS_TYPE
          bi.bitmap_ptrs = blk.bitmap_ptrs
          bi.next_blk = blk.bitmap_ext_blk
        # --- root block ---
        elif blk.is_root_block():
          bi.blk_type = self.BT_ROOT
          bi.blk_status = self.BS_TYPE
          root = RootBlock(self.blkdev, blk_num)
          root.set(data)
          bi.name = FSString(root.name)
          bi.hash_table = root.hash_table
          bi.parent_blk = 0
          self.log.msg(Log.DEBUG, "Found Root: '%s'" % bi.name, blk_num)
          # chech hash size
          nht = len(root.hash_table)
          if root.hash_size != nht:
            self.log.msg(Log.ERROR, "Root block hash table size mismatch", blk_num)
          eht = self.blkdev.block_longs - 56
          if nht != eht:
            self.log.msg(Log.WARN, "Root block does not have normal hash size: %d != %d" % (nht, eht), blk_num)
        # --- user dir block ---
        elif blk.is_user_dir_block():
          bi.blk_type = self.BT_DIR
          bi.blk_status = self.BS_TYPE
          user = UserDirBlock(self.blkdev, blk_num, DosType.is_longname(self.dos_type))
          user.set(data)
          bi.name = FSString(user.name)
          bi.parent_blk = user.parent
          bi.next_blk = user.hash_chain
          bi.hash_table = user.hash_table
          bi.own_key = user.own_key
          self.log.msg(Log.DEBUG, "Found Dir : '%s'" % bi.name, blk_num)
        # --- filter header block ---
        elif blk.is_file_header_block():
          bi.blk_type = self.BT_FILE_HDR
          bi.blk_status = self.BS_TYPE
          fh = FileHeaderBlock(self.blkdev, blk_num, DosType.is_longname(self.dos_type))
          fh.set(data)
          bi.name = FSString(fh.name)
          bi.parent_blk = fh.parent
          bi.next_blk = fh.hash_chain
          bi.own_key = fh.own_key
          bi.byte_size = fh.byte_size
          bi.data_blocks = fh.data_blocks
          bi.extension = fh.extension
          self.log.msg(Log.DEBUG, "Found File: '%s'" % bi.name, blk_num)
        # --- file list block ---
        elif blk.is_file_list_block():
          bi.blk_type = self.BT_FILE_LIST
          fl = FileListBlock(self.blkdev, blk_num)
          fl.set(data)
          bi.blk_status = self.BS_TYPE
          bi.ext_blk = fl.extension
          bi.blk_list = fl.data_blocks
          bi.own_key = fl.own_key
          bi.data_blocks = fl.data_blocks
          bi.extension = fl.extension
          bi.parent_blk = fl.parent
        # --- file data block (OFS) ---
        elif blk.is_file_data_block():
          bi.blk_type = self.BT_FILE_DATA
          bi.blk_status = self.BS_TYPE
          fd = FileDataBlock(self.blkdev, blk_num)
          fd.set(data)
          bi.data_size = fd.data_size
          bi.hdr_key = fd.hdr_key
          bi.seq_num = fd.seq_num
        elif blk.is_comment_block():
          bi.blk_type = self.BT_COMMENT
          bi.blk_status = self.BS_TYPE
          cblk = CommentBlock(self.blkdev, blk_num)
          bi.hdr_key = cblk.hdr_key
          bi.own_key = cblk.own_key
              
    except IOError as e:
      self.log.msg(Log.ERROR, "Can't read block", blk_num)
      bi = BlockInfo(blk_num)
      bi.blk_status = BS_READ_ERROR
      
    # sort block info into map and arrays assigned by status/type
    self.block_map[blk_num] = bi
    self.map_status[bi.blk_status].append(bi)
    self.map_type[bi.blk_type].append(bi)
    return bi
    
  def any_chance_of_fs(self):
    """is there any chance to find a FS on this block device?"""
    num_dirs = len(self.map_type[self.BT_DIR])
    num_files = len(self.map_type[self.BT_FILE_HDR])
    num_roots = len(self.map_type[self.BT_ROOT])
    return (num_files > 0) or ((num_roots + num_dirs) > 0)
    
  def get_blocks_of_type(self, t):
    return self.map_type[t]

  def get_blocks_with_key_value(self, key, value):
    res = []
    for bi in self.block_map:
      if hasattr(bi, key):
        v = getattr(bi, key)
        if v == value:
          res.append(bi)
    return res

  def is_block_available(self, num):
    if num >= 0 and num < len(self.block_map):
      return self.block_map[num] != None
    else:
      return False

  def get_block(self, num):
    if num >= 0 and num < len(self.block_map):
      bi = self.block_map[num]
      if bi == None:
        bi = self.read_block(num)
      return bi
    else:
      return None

  def dump(self):
    for b in self.block_map:
      if b != None:
        print(b)




import hashlib
import os
from binascii import hexlify

from fsbc import settings
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.drivers.messdriver import MessDriver
from fsgs.drivers.retroarchdriver import RetroArchDriver
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

SMD_MODEL_NTSC = "ntsc"
SMD_MODEL_NTSC_J = "ntsc-j"
SMD_MODEL_PAL = "pal"
SMD_CONTROLLER_TYPE = "gamepad"
SMD_CONTROLLER = {
    "type": SMD_CONTROLLER_TYPE,
    "description": "Gamepad",
    "mapping_name": "megadrive",
}
SMD_6BUTTON_CONTROLLER_TYPE = "gamepad6"
SMD_6BUTTON_CONTROLLER = {
    "type": SMD_6BUTTON_CONTROLLER_TYPE,
    "description": "Gamepad",
    "mapping_name": "megadrive",
}
NO_CONTROLLER_TYPE = "none"
NO_CONTROLLER = {
    "type": NO_CONTROLLER_TYPE,
    "description": "None",
    "mapping_name": "",
}
SMD_PORTS = [
    {
        "description": "Port 1",
        "types": [SMD_CONTROLLER, SMD_6BUTTON_CONTROLLER, NO_CONTROLLER],
        "type_option": "smd_port_1_type",
        "device_option": "smd_port_1",
    },
    {
        "description": "Port 2",
        "types": [SMD_CONTROLLER, SMD_6BUTTON_CONTROLLER, NO_CONTROLLER],
        "type_option": "smd_port_2_type",
        "device_option": "smd_port_2",
    },
]


class MegaDrivePlatform(Platform):
    PLATFORM_NAME = "Mega Drive"

    def driver(self, fsgc):
        emulator = settings.get(Option.SMD_EMULATOR)

        if emulator in ["retroarch-fs", "retroarch-fs/genesisplusgx"]:
            return MegaDriveRetroArchDriver(fsgc)
        if emulator in ["mame-fs"]:
            return MegaDriveMameDriver(fsgc)
        if emulator in ["mednafen"]:
            return MegaDriveMednafenDriver(fsgc)

        # FIXME: Vanilla retroarch not supported yet
        if emulator in ["retroarch", "retroarch/genesisplusgx"]:
            return MegaDriveRetroArchDriver(fsgc)
        # Deprecated name
        if emulator in ["retroarch-genesis-plus-gx"]:
            return MegaDriveRetroArchDriver(fsgc)
        # Legacy names
        if emulator in ["mame", "mess"]:
            return MegaDriveMameDriver(fsgc)

        return MegaDriveMednafenFSDriver(fsgc)

    def loader(self, fsgc):
        return MegaDriveLoader(fsgc)


class MegaDriveLoader(SimpleLoader):
    def load(self, values):
        super().load(values)
        self.config[Option.SMD_MODEL] = values["smd_model"]
        self.config[Option.SMD_PORT_1_TYPE] = values["smd_port_1_type"]
        self.config[Option.SMD_PORT_2_TYPE] = values["smd_port_2_type"]
        # self.config[Option.SMD_PORT_3_TYPE] = values["smd_port_3_type"]
        # self.config[Option.SMD_PORT_4_TYPE] = values["smd_port_4_type"]

        # FIXME: Should be able to remove this now (smd_model in database)
        if not self.config[Option.SMD_MODEL]:
            variant = values["variant_name"].lower()
            if "world" in variant or "usa" in variant:
                model = SMD_MODEL_NTSC
            elif "europe" in variant or "australia" in variant:
                model = SMD_MODEL_PAL
            elif "japan" in variant:
                model = SMD_MODEL_NTSC_J
            else:
                # FIXME: Remove?
                # model = SMD_MODEL_AUTO
                model = SMD_MODEL_NTSC
            self.config[Option.SMD_MODEL] = model


class MegaDriveMameDriver(MessDriver):
    PORTS = SMD_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = MegaDriveHelper(self.options)
        self.save_handler.set_save_data_is_emulator_specific(True)
        self.save_handler.set_mame_driver(self.get_mame_driver())

    def prepare(self):
        print("[SMD] MAME driver preparing...")
        super().prepare()
        self.emulator.args.extend(["-cart", self.helper.prepare_rom(self)])

    def get_mame_driver(self):
        self.helper.set_model_name_from_model(self)
        if self.helper.model() == SMD_MODEL_NTSC:
            return "genesis"
        elif self.helper.model() == SMD_MODEL_PAL:
            return "megadriv"
        elif self.helper.model() == SMD_MODEL_NTSC_J:
            return "megadrij"
        raise Exception("Could not determine SMD MAME driver")

    def get_mess_input_mapping(self, _):
        return {
            "START": "P#_START",
            "MODE": "P#_SELECT",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "A": "P#_BUTTON1",
            "B": "P#_BUTTON2",
            "C": "P#_BUTTON3",
            "X": "P#_BUTTON4",
            "Y": "P#_BUTTON5",
            "Z": "P#_BUTTON6",
        }

    def get_mess_romset(self):
        return self.get_mame_driver(), {}


class MegaDriveMednafenDriver(MednafenDriver):
    PORTS = SMD_PORTS

    def __init__(self, fsgc, vanilla=True):
        super().__init__(fsgc, vanilla)
        self.helper = MegaDriveHelper(self.options)
        self.save_handler.set_save_data_is_emulator_specific(True)
        # self.save_handler.set_srm_alias(".sav")

    def prepare(self):
        print("[SMD] Mednafen driver preparing...")
        super().prepare()
        rom_path = self.helper.prepare_rom(self)
        self.helper.read_rom_metadata(rom_path)
        self.init_mega_drive_model()
        for i in range(len(self.PORTS)):
            self.init_mega_drive_port(i)
        self.init_mednafen_crop_from_viewport()
        self.set_mednafen_aspect(4, 3)
        # We do aspect calculation separately. Must not be done twice.
        self.emulator.args.extend(["-md.correct_aspect", "0"])
        # ROM path must be added at the end of the argument list
        self.emulator.args.append(rom_path)

    def init_mega_drive_model(self):
        self.helper.set_model_name_from_model(self)
        if self.helper.model() == SMD_MODEL_NTSC:
            region = "overseas_ntsc"
        elif self.helper.model() == SMD_MODEL_PAL:
            region = "overseas_pal"
        elif self.helper.model() == SMD_MODEL_NTSC_J:
            region = "domestic_ntsc"
        else:
            # FIXME: This model might disappear
            region = "game"
        self.emulator.args.extend(["-md.region", region])

    def init_mega_drive_port(self, i):
        t = self.ports[i].type
        n = i + 1
        if t == SMD_CONTROLLER_TYPE:
            t = "gamepad"
        elif t == SMD_6BUTTON_CONTROLLER_TYPE:
            t = "gamepad6"
        elif t == NO_CONTROLLER_TYPE:
            t = "none"
        else:
            self.logger.warning(
                "Unknown port type '{}' for Mega Drive port {}".format(t, n)
            )
            t = "gamepad"
        self.emulator.args.extend(["-md.input.port{}".format(n), t])

    def init_mednafen_crop_from_viewport(self):
        viewport = self.options[Option.VIEWPORT]
        if viewport:
            if viewport.startswith("0 0 320 240 ="):
                viewport_src, viewport = self.mednafen_viewport()
                self.emulator.env["FSGS_CROP"] = "{},{},{},{}".format(
                    viewport[0], viewport[1], viewport[2], viewport[3]
                )
        else:
            if self.scaling == self.NO_SCALING:
                if self.helper.model() == SMD_MODEL_PAL:
                    self.emulator.env["FSGS_CROP"] = "0,0,320,240"
                else:
                    self.emulator.env["FSGS_CROP"] = "0,8,320,224"

    def mednafen_system_prefix(self):
        return "md"

    def mednafen_input_mapping(self, port):
        n = port + 1
        return {
            "A": "md.input.port{}.gamepad.a".format(n),
            "B": "md.input.port{}.gamepad.b".format(n),
            "C": "md.input.port{}.gamepad.c".format(n),
            "X": "md.input.port{}.gamepad.x".format(n),
            "Y": "md.input.port{}.gamepad.y".format(n),
            "Z": "md.input.port{}.gamepad.z".format(n),
            "UP": "md.input.port{}.gamepad.up".format(n),
            "DOWN": "md.input.port{}.gamepad.down".format(n),
            "LEFT": "md.input.port{}.gamepad.left".format(n),
            "RIGHT": "md.input.port{}.gamepad.right".format(n),
            "MODE": "md.input.port{}.gamepad.mode".format(n),
            "START": "md.input.port{}.gamepad.start".format(n),
        }

        # FIXME: add PAUSE button to universal gamepad config

    # def game_video_par(self):
    #     # These may not be entirely correct...
    #     if self.is_pal():
    #         return (4 / 3) / (320 / 240)
    #     else:
    #         return (4 / 3) / (320 / 224)

    def game_video_size(self):
        viewport_src, viewport = self.mednafen_viewport()
        if viewport is not None:
            return viewport[2], viewport[3]
        if self.helper.model() == SMD_MODEL_PAL:
            return 320, 240
        else:
            return 320, 224

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class MegaDriveMednafenFSDriver(MegaDriveMednafenDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs, vanilla=False)


class MegaDriveRetroArchDriver(RetroArchDriver):
    PORTS = SMD_PORTS

    def __init__(self, fsgc):
        super().__init__(
            fsgc, "genesis_plus_gx_libretro", "RetroArch/GenesisPlusGX"
        )
        self.helper = MegaDriveHelper(self.options)
        self.save_handler.set_save_data_is_emulator_specific(True)

    def prepare(self):
        print("[SMD] RetroArch/GenesisPlusGX driver preparing...")
        super().prepare()
        hw = "mega drive / genesis"
        region = self.init_mega_drive_model()
        for i in range(len(self.PORTS)):
            self.init_mega_drive_port(i)
        with self.open_retroarch_core_options() as f:
            f.write('genesis_plus_gx_system_hw = "{}"\n'.format(hw))
            f.write("genesis_plus_gx_region_detect = {}\n".format(region))
        self.emulator.args.append(self.helper.prepare_rom(self))

    def init_mega_drive_model(self):
        self.helper.set_model_name_from_model(self)
        if self.helper.model() == SMD_MODEL_NTSC:
            return "ntsc-u"
        elif self.helper.model() == SMD_MODEL_PAL:
            return "pal"
        elif self.helper.model() == SMD_MODEL_NTSC_J:
            return "ntsc-j"
        return "ntsc-u"

    def init_mega_drive_port(self, i):
        t = self.ports[i].type
        n = i + 1
        if t == SMD_CONTROLLER_TYPE:
            t = "257"
        elif t == SMD_6BUTTON_CONTROLLER_TYPE:
            t = "513"
        elif t == NO_CONTROLLER_TYPE:
            t = "0"
        else:
            self.logger.warning(
                "Unknown port type '{}' for Mega Drive port {}".format(t, n)
            )
            t = "gamepad"
        # FIXME: Note, this only works for p1 and p2. Port 3 and beyond
        # supports RetroPad only?
        self.retroarch_config["input_libretro_device_p{}".format(n)] = t

    def game_video_size(self):
        # # FIXME: Account for horizontal overscan
        # # FIXME: Account for vertical overscan
        #
        # viewport = self.options[Option.VIEWPORT]
        # if viewport == "0 0 256 240 = 0 0 256 240":
        #     return 256, 240
        # # elif viewport == "0 0 256 240 = 0 8 256 224":
        # #     return 256, 224
        # elif viewport == "0 0 256 240 = 8 0 240 240":
        #     return 240, 240
        # elif viewport == "0 0 256 240 = 8 8 240 224":
        #     return 240, 224
        # else:
        #     return 256, 224

        # FIXME: PAL?
        return 320, 224

    def retroarch_input_mapping(self, port):
        n = port + 1
        return {
            "A": "input_player{}_y".format(n),
            "B": "input_player{}_b".format(n),
            "C": "input_player{}_a".format(n),
            "X": "input_player{}_l".format(n),
            "Y": "input_player{}_x".format(n),
            "Z": "input_player{}_r".format(n),
            "UP": "input_player{}_up".format(n),
            "DOWN": "input_player{}_down".format(n),
            "LEFT": "input_player{}_left".format(n),
            "RIGHT": "input_player{}_right".format(n),
            "MODE": "input_player{}_select".format(n),
            "START": "input_player{}_start".format(n),
        }

    # def window_size(self):
    #     return 256, 224


class MegaDriveHelper:
    def __init__(self, options):
        self.options = options

    def model(self):
        if self.options[Option.SMD_MODEL] == SMD_MODEL_NTSC:
            return SMD_MODEL_NTSC
        if self.options[Option.SMD_MODEL] == SMD_MODEL_NTSC_J:
            return SMD_MODEL_NTSC_J
        if self.options[Option.SMD_MODEL] == SMD_MODEL_PAL:
            return SMD_MODEL_PAL
        # FIXME: REMOVE?
        # return SMD_MODEL_AUTO
        return SMD_MODEL_NTSC

    def set_model_name_from_model(self, driver):
        model = self.model()
        if model == SMD_MODEL_NTSC:
            driver.set_model_name("Genesis")
        elif model == SMD_MODEL_PAL:
            driver.set_model_name("Mega Drive PAL")
        elif model == SMD_MODEL_NTSC_J:
            driver.set_model_name("Mega Drive NTSC-J")
        # else:
        #     # FIXME: This model might disappear
        #     # driver.set_model_name("Mega Drive / Genesis")
        #     assert False

    def read_rom_metadata(self, rom_path):
        # FIXME: Move to MegaDriveHeaderParser
        # http://md.squee.co/Howto:Initialise_a_Mega_Drive
        # https://en.wikibooks.org/wiki/Genesis_Programming
        with open(rom_path, "rb") as f:
            data = f.read(0x200)
        if not len(data) == 0x200:
            print("[SMD] Did not read 0x200 header bytes")
            return
        region = data[0x200 - 16 :]
        # notes = data[0x200 - 16 - 52]
        sram_end = data[0x200 - 16 - 52 - 4 * 1 : 0x200 - 16 - 52 - 4 * 0]
        sram_start = data[0x200 - 16 - 52 - 4 * 2 : 0x200 - 16 - 52 - 4 * 1]
        sram_type = data[0x200 - 16 - 52 - 4 * 3 : 0x200 - 16 - 52 - 4 * 2]

        sram_end = hexlify(sram_end)
        sram_start = hexlify(sram_start)
        # sram_type = hexlify(sram_type)
        sram_start = int(sram_start, 16)
        sram_end = int(sram_end, 16)
        print("[SMD] Region:", repr(region))
        print("[SMD] SRAM Type :", repr(sram_type))
        if sram_type[0] == 0x52:
            print(" - 0x52 OK")
        else:
            print(" - 0x52 MISSING")
        if sram_type[1] == 0x41:
            print(" - 0x41 OK")
        else:
            print(" - 0x41 MISSING")
        is_backup = (sram_type[2] & 0b01000000) != 0
        odd, even = False, False
        if (sram_type[2] & 0b00011000) == 0b00011000:
            odd = True
        elif (sram_type[2] & 0b00011000) == 0b00010000:
            even = True
        elif (sram_type[2] & 0b00011000) == 0b00000000:
            even = True
            odd = True
        print(" - Is backup RAM?", is_backup)
        print(" - Odd?", odd, "Even?", even)
        print("[SMD] SRAM Start:", repr(sram_start))
        print("[SMD] SRAM End  :", repr(sram_end))

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom_with_stream(self, driver, input_stream, ext):
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path


class MegaDriveHeaderParser:
    def __init__(self):
        self.sram = False
        self.sram_start = 0
        self.sram_end = 0
        self.sram_size = 0
        self.sram_odd_only = False
        self.sram_even_only = False
        self.sram_odd_and_even = False

    def parse_file(self, path):
        with open(path, "rb") as f:
            data = f.read(0x200)
        if not len(data) == 0x200:
            return False
        return self.parse_data(data)

    def parse_data(self, data):
        if not len(data) >= 0x200:
            return False



import time

import requests

import fsui
from fsbc.application import app
from fsbc.signal import Signal
from fsbc.task import Task
from fsgs.Database import Database
from fsgs.download import Downloader
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.ogd.context import SynchronizerContext
from fsui.extra.iconheader import IconHeader
from launcher.filescanner import FileScanner
from launcher.gamescanner import GameScanner

from ..i18n import gettext


class DownloadGameWindow(fsui.Window):
    def __init__(self, parent, fsgs):
        fsui.Window.__init__(self, parent, gettext("Download Game Manually"))
        self.fsgs = fsgs
        self.download_page = fsgs.config.get("download_page")
        self.download_notice = fsgs.config.get("download_notice")
        self.task = None

        self.layout = fsui.VerticalLayout()
        self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = IconHeader(
            self,
            fsui.Icon("web-browser", "pkg:workspace"),
            gettext("Download Game Manually"),
            gettext("This game must be downloaded before you can play it"),
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        label = fsui.HeadingLabel(
            self,
            gettext(
                "Please open the following web page and download the "
                "game from there:"
            ),
        )
        label.set_min_width(500)
        self.layout.add(label)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=10)

        hori_layout.add(
            fsui.ImageView(
                self, fsui.Image("launcher:res/16x16/world_link.png")
            )
        )

        label = fsui.URLLabel(self, self.download_page, self.download_page)
        hori_layout.add(label, margin_left=6)

        if self.download_notice:
            label = fsui.MultiLineLabel(self, self.download_notice)
            label.set_min_width(500)
            self.layout.add(label, margin_top=20)

        label = fsui.HeadingLabel(
            self,
            gettext(
                "Download to the following directory, and then "
                "click '{0}':".format(gettext("Scan Downloads"))
            ),
        )
        label.set_min_width(500)
        self.layout.add(label, margin_top=20)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=10)

        hori_layout.add(
            fsui.ImageView(self, fsui.Image("launcher:res/16x16/folder.png"))
        )

        label = fsui.Label(self, FSGSDirectories.ensure_downloads_dir())
        hori_layout.add(label, margin_left=6)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=20)

        self.status_label = fsui.Label(self, "")
        hori_layout.add(self.status_label, expand=True)

        self.scan_button = fsui.Button(self, gettext("Scan Downloads"))
        self.scan_button.activated.connect(self.on_scan_files)
        hori_layout.add(self.scan_button, margin_left=20)

        self.close_button = fsui.Button(self, gettext("Close"))
        self.close_button.activated.connect(self.on_close)
        hori_layout.add(self.close_button, margin_left=10)

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

    def on_close(self):
        self.close()

    def on_scan_files(self):
        self.scan_button.disable()

        self.task = RescanTask()
        self.task.progressed.connect(self.on_progress)
        self.task.failed.connect(self.on_failure)
        self.task.succeeded.connect(self.on_success)
        # self.task.stopped.connect(self.close)
        self.task.start()

    def on_progress(self, message):
        if not isinstance(message, str):
            message = message[0]
        # self.icon_header.subtitle_label.set_text(message)
        self.status_label.set_text(message)

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())
        self.scan_button.enable()
        self.status_label.set_text("")

    def on_success(self):
        if self.fsgs.config.get("x_missing_files"):
            message = gettext(
                "Files for this game are still missing. Did you download "
                "the game and put the file(s) in the Downloads directory?"
            )
            fsui.show_error(
                message,
                title=gettext("Missing Files"),
                parent=self.get_window(),
            )
            self.scan_button.enable()
            self.status_label.set_text("")
        else:
            self.close()


class RescanTask(Task):
    def __init__(self):
        Task.__init__(self, "RescanTask")

    # def stop_check(self):
    #     pass

    def on_status(self, status):
        self.progressed(status)

    def run(self):
        with Database.get_instance() as database:
            self._run(database)

    def _run(self, database):
        context = SynchronizerContext()

        paths = [FSGSDirectories.downloads_dir()]
        scanner = FileScanner(
            paths,
            purge_other_dirs=False,
            on_status=self.on_status,
            stop_check=self.stop_check,
        )
        scanner.scan()

        scanner = GameScanner(
            context, None, on_status=self.on_status, stop_check=self.stop_check
        )
        scanner.scan(database)

        # FIXME: review what signals should be sent when a scan is performed
        # FIXME: these should be removed soon
        app.settings["last_scan"] = str(time.time())
        app.settings["__config_refresh"] = str(time.time())
        # this must be called from main, since callbacks are broadcast
        # when settings are changed
        Signal("scan_done").notify()


class DownloadTermsDialog(fsui.LegacyDialog):
    def __init__(self, parent, fsgs):
        fsui.LegacyDialog.__init__(self, parent, gettext("Terms of Download"))
        self.fsgs = fsgs
        self.download_file = fsgs.config.get("download_file")
        self.download_terms = fsgs.config.get("download_terms")
        self.task = None

        self.layout = fsui.VerticalLayout()
        self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = IconHeader(
            self,
            fsui.Icon("web-browser", "pkg:workspace"),
            gettext("Terms of Download"),
            gettext(
                "This game can be automatically downloaded if you "
                "accept the terms"
            ),
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        self.label = fsui.MultiLineLabel(self, gettext("Loading..."))
        self.label.set_min_width(600)
        self.label.set_min_height(100)
        self.layout.add(self.label)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=20)

        hori_layout.add_spacer(0, expand=True)

        self.reject_button = fsui.Button(self, gettext("Reject"))
        self.reject_button.activated.connect(self.on_reject_button)
        hori_layout.add(self.reject_button, margin_left=10)

        self.accept_button = fsui.Button(self, gettext("Accept"))
        self.accept_button.activated.connect(self.on_accept_button)
        hori_layout.add(self.accept_button, margin_left=10)

        self.accept_button.disable()
        self.reject_button.disable()

        self.task = DownloadTermsTask(self.download_terms)
        # self.task.progressed.connect(self.on_progress)
        self.task.failed.connect(self.on_failure)
        self.task.succeeded.connect(self.on_success)
        self.task.start()

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

    def on_accept_button(self):
        self.end_modal(True)
        Downloader.set_terms_accepted(self.download_file, self.download_terms)

    def on_reject_button(self):
        self.end_modal(False)

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())

    def on_success(self):
        self.label.set_text(self.task.data)
        self.reject_button.enable()
        self.accept_button.enable()
        self.accept_button.focus()


class DownloadTermsTask(Task):
    def __init__(self, url):
        Task.__init__(self, "DownloadTermsTask")
        self.url = url
        self.data = None

    def run(self):
        for i in range(3):
            try:
                r = requests.get(self.url)
                r.raise_for_status()
                self.data = r.text
            except Exception:
                time.sleep(0.5)
            else:
                return
        r = requests.get(self.url)
        r.raise_for_status()
        self.data = r.text



from __future__ import absolute_import
from __future__ import print_function

import subprocess
import tempfile
import os

class DisAsm:
  def __init__(self, use_objdump = False, cpu = '68000'):
    self.use_objdump = use_objdump
    self.cpu = cpu
 
  def _parse_vda68k(self, lines):
    # parse output: split addr, raw words, and code
    result = []
    for l in lines:
      addr = int(l[0:8],16)
      word = map(lambda x: int(x,16),l[10:30].split())      
      code = l[30:]
      result.append((addr,word,code))
    return result
 
  def _parse_objdump(self, lines):
    result = []
    for l in lines[7:]:
      if not '...' in l:
        addr_str = l[:8].strip()
        word_str = l[9:26].split()
        code = l[26:]
      
        addr = int(addr_str,16)
        word = map(lambda x: int(x,16),word_str)
        result.append((addr,word,code))
    return result
 
  def disassemble(self, data, start=0):
    # write to temp file
    tmpname = tempfile.mktemp()
    out = file(tmpname,"wb")
    out.write(data)
    out.close()
    
    if self.use_objdump:
      cmd = ["m68k-elf-objdump","-D","-b","binary","-m",self.cpu,tmpname]
    else:
      cmd = ["vda68k",tmpname,str(start)]
    
    # call external disassembler
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output = p.communicate()[0]
    os.remove(tmpname)
    lines = output.splitlines()
    
    if self.use_objdump:
      return self._parse_objdump(lines)
    else:
      return self._parse_vda68k(lines)

  def dump(self, code):
    for line in code:
      ops = map(lambda x : "%04x" % x, line[1])
      print("%08x:  %-20s  %s" % (line[0]," ".join(ops),line[2]))



import os
from functools import lru_cache

from fsbc.application import app
from fsbc.settings import Settings
from fsgs.FSGSDirectories import FSGSDirectories


@lru_cache()
def default_openretro_server_from_file():
    server = None
    p = os.path.join(
        FSGSDirectories.get_data_dir(), "Settings", "database-server"
    )
    if os.path.exists(p):
        with open(p, "r", encoding="UTF-8") as f:
            server = f.read().strip()
    return server


def default_openretro_server():
    return "https://openretro.org"


def fs_uae_url_from_sha1_and_name(sha1, name):
    return "https://fs-uae.net/s/sha1/{0}/{1}".format(sha1, name)


def openretro_server():
    server = Settings.instance()["database_server"]
    if not server:
        server = default_openretro_server_from_file()
    if not server:
        server = default_openretro_server()
    if "://" in server:
        scheme, host = server.split("://")
    else:
        scheme = "http"
        host = server
    return scheme, host


def openretro_scheme():
    # if openretro_host() == default_openretro_host():
    #     return "https"
    return "http"


def openretro_url_prefix():
    scheme, host = openretro_server()
    return "{}://{}".format(scheme, host)


def is_http_url(url):
    return url.startswith("http://") or url.startswith("https://")



from __future__ import absolute_import
from __future__ import print_function

from .BlockDevice import BlockDevice
from .DiskGeometry import DiskGeometry
from .ImageFile import ImageFile
import os.path
import os

class HDFBlockDevice(BlockDevice):
  def __init__(self, hdf_file, read_only=False, block_size=512, fobj=None):
    self.img_file = ImageFile(hdf_file, read_only, block_size, fobj)

  def create(self, geo, reserved=2):
    self._set_geometry(geo.cyls, geo.heads, geo.secs, reserved=reserved)
    self.img_file.create(geo.get_num_blocks())
    self.img_file.open()

  def open(self, geo, reserved=2):
    self._set_geometry(geo.cyls, geo.heads, geo.secs, reserved=reserved)
    self.img_file.open()

  def flush(self):
    pass

  def close(self):
    self.img_file.close()

  def read_block(self, blk_num):
    return self.img_file.read_blk(blk_num)

  def write_block(self, blk_num, data):
    return self.img_file.write_blk(blk_num, data)



import fsui
from fsgs.plugins.pluginmanager import PluginManager
from launcher.res import gettext
from launcher.settings.settings_page import SettingsPage


class PluginsSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("plugins", "pkg:workspace")
        self.add_header(icon, gettext("Plugins"))

        self.list_view = fsui.ListView(self)
        self.list_view.set_min_height(140)
        # self.list_view.item_activated.connect(self.on_plugin_activated)
        image = fsui.Image("workspace:res/16x16/settings.png")
        plugin_manager = PluginManager.instance()
        for plugin in plugin_manager.plugins():
            text = "{0} ({1})".format(plugin.name, plugin.version)
            if plugin.outdated:
                text += " - " + gettext("This plugin is outdated").upper()
            print("[PLUGINS] {0}".format(text))
            self.list_view.add_item(text, icon=image)
        self.layout.add(self.list_view, fill=True, expand=True)



import os
import time

import fsgs
from arcade.gamecentersettings import GameCenterSettings
from arcade.glui.constants import TOP_ITEM_LEFT, TOP_ITEM_NOBORDER
from arcade.glui.font import BitmapFont
from arcade.glui.items import MenuItem, create_item_menu
from arcade.glui.navigatable import Navigatable
from arcade.glui.opengl import fs_emu_blending
from OpenGL import GL as gl
from arcade.glui.sdl import *
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.resources import gettext
from fsbc.application import app
from fsbc.settings import Settings


def post_quit_event():
    State.get().quit = True


class TopMenuItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.normal_texture = None
        self.selected_texture = None

    def update_size_right(self):
        self.w = 80

    def render_top_right(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            style=TOP_ITEM_LEFT,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        fs_emu_blending(True)
        if selected:
            texture = self.selected_texture
        else:
            texture = self.normal_texture
        # texture.render(self.x, self.y, self.w, self.h)
        texture.render(
            self.x + (self.w - texture.w) / 2,
            self.y + (self.h - texture.h) / 2,
            texture.w,
            texture.h,
        )


class CloseItem(TopMenuItem):
    def __init__(self):
        super().__init__()
        self.normal_texture = Texture.close
        self.selected_texture = Texture.close_selected

    def activate(self, menu):
        post_quit_event()


class ShutdownItem(TopMenuItem):
    def __init__(self):
        super().__init__()
        self.normal_texture = Texture.shutdown
        self.selected_texture = Texture.shutdown_selected

    def activate(self, menu):
        command = GameCenterSettings.get_shutdown_command()
        if command:
            os.system(command)
        post_quit_event()


class MaximizeItem(TopMenuItem):
    def __init__(self):
        super().__init__()
        self.title = "Ma"


class MinimizeItem(TopMenuItem):
    def __init__(self):
        super().__init__()
        self.normal_texture = Texture.minimize
        self.selected_texture = Texture.minimize_selected

    def activate(self, menu):
        SDL_Minimize()
        # want item menu focused when window is restored
        # FIXME: temporarily disabled
        # set_current_menu(State.get().current_menu)


class ClockItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.title = "00:00"
        self.enabled = False

    def update_size(self, text):
        self.w = 148

    def render_top_right(self, selected=False):
        self.render_top_background(selected, style=TOP_ITEM_NOBORDER)
        self.render_top(self.get_top_right_text(), selected, right_align=True)

    def get_top_right_text(self):
        return time.strftime("%H:%M")

    def activate(self, menu):
        pass


class AspectItem(TopMenuItem):
    def __init__(self):
        super().__init__()
        # self.normal_texture = Texture.stretch
        # self.selected_texture = Texture.stretch
        self.normal_texture = Texture.aspect
        self.selected_texture = Texture.aspect
        self.update_texture()

    # def update_texture(self):
    #     # TODO: Ideally, this class should listen for settings changes.
    #     if Settings.instance()["keep_aspect"] in ["0", ""]:
    #         texture = Texture.stretch
    #     else:
    #         texture = Texture.aspect
    #     self.normal_texture = texture
    #     self.selected_texture = texture
    #
    # def activate(self, menu):
    #     if Settings.instance()["keep_aspect"] == "1":
    #         Settings.instance()["keep_aspect"] = ""
    #     else:
    #         Settings.instance()["keep_aspect"] = "1"
    #     self.update_texture()

    def update_texture(self):
        # TODO: Ideally, this class should listen for settings changes.
        if Settings.instance()["stretch"] == "1":
            texture = Texture.stretch
        elif Settings.instance()["stretch"] == "0":
            texture = Texture.square_pixels
        else:
            texture = Texture.aspect
        self.normal_texture = texture
        self.selected_texture = texture

    def activate(self, menu):
        if Settings.instance()["stretch"] == "1":
            Settings.instance()["stretch"] = "0"
        elif Settings.instance()["stretch"] == "0":
            Settings.instance()["stretch"] = ""
        else:
            Settings.instance()["stretch"] = "1"
        self.update_texture()


class VideoSyncItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.title = "V-SYNC"

    def activate(self, menu):
        if Settings.instance()["video_sync"] == "1":
            Settings.instance()["video_sync"] = ""
        else:
            Settings.instance()["video_sync"] = "1"

    def render_top_right(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            style=TOP_ITEM_LEFT,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_blending(True)
        if Settings.instance()["video_sync"] == "1":
            r = 1.0
            g = 1.0
            b = 1.0
            alpha = 1.0
        else:
            r = 1.0
            g = 1.0
            b = 1.0
            alpha = 0.33
        x = self.x + 20
        BitmapFont.title_font.render(
            self.title, x, self.y + 14, r=r, g=g, b=b, alpha=alpha
        )
        gl.glEnable(gl.GL_DEPTH_TEST)


class TopMenu(Navigatable):
    def __init__(self):
        super().__init__()
        self.left = []
        self.right = []
        self.right.append(VideoSyncItem())
        self.right.append(AspectItem())
        if self.use_clock_item():
            self.right.append(ClockItem())
        # if Render.get().allow_minimize:
        #     if Config.get_bool("menu/minimize", True):
        #         self.right.append(MinimizeItem())
        command = GameCenterSettings.get_shutdown_command()
        if command:
            self.right.append(ShutdownItem())
        else:
            self.right.append(CloseItem())
        self._selected_index = 0

    @staticmethod
    def use_clock_item():
        return app.settings["game-center:top-clock"] != "0"

    def append_left(self, item):
        self.left.append(item)

    def append_right(self, item):
        self.right.insert(0, item)

    @property
    def selected_item(self):
        index = self._selected_index
        return self[index]

    def get_selected_index(self):
        return self._selected_index

    def set_selected_index(self, index, immediate=False):
        assert immediate is not None
        self._selected_index = index

    def go_left(self, count=1):
        index = self._selected_index - 1
        while index >= 0:
            if self[index].enabled:
                self._selected_index = index
                break
            index -= 1

    def go_right(self, count=1):
        index = min(len(self) - 1, self._selected_index + count)
        self._selected_index = index
        if not self[index].enabled:
            self.go_right()

    def go_up(self):
        # do nothing
        pass

    def activate(self):
        result = self.selected_item.activate(State.get().current_menu)
        # FIXME:
        from arcade.glui.menu import Menu

        if isinstance(result, Menu):
            # FIXME:
            from arcade.glui.window import enter_menu

            enter_menu(result)

    def __getitem__(self, index):
        if index < len(self.left):
            return self.left[index]
        return self.right[index - len(self.left)]

    def __len__(self):
        return len(self.left) + len(self.right)


class OldGameCenterItem(MenuItem):
    def __init__(self):
        super().__init__()
        if fsgs.product == "FS-UAE":
            self.title = "FS-UAE   Arcade"
        elif fsgs.product == "OpenRetro":
            self.title = "OpenRetro   Arcade"
        else:
            self.title = "Unknown   Arcade"
        # else:
        #     self.title = gettext("Game   Center")
        self.path_title = self.title
        self.enabled = False

    def activate(self, menu):
        pass

    # def update_size(self, text):
    #     MenuItem.update_size(text)
    #     #self.w = Texture.top_logo.w + 83

    def render_top_left(self, selected=False):
        # self.render_top_background(selected, style=TOP_ITEM_ARROW)
        MenuItem.render_top_left(self, selected=selected)
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_blending(True)
        # if app.name == "fs-uae-arcade":
        x = 161
        # else:
        #     x = 138
        y = 14
        # if selected:
        # texture = Texture.top_logo_selected
        # else:
        texture = Texture.top_logo
        texture.render(x, 1080 - y - texture.h, texture.w, texture.h)
        # fs_emu_blending(False)
        gl.glEnable(gl.GL_DEPTH_TEST)


class GameCenterItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.path_title = self.title
        self.enabled = False
        if fsgs.product == "FS-UAE":
            self.text1 = "FS-UAE"
        elif fsgs.product == "OpenRetro":
            self.text1 = "OpenRetro"
        else:
            self.text1 = "Unknown"
        self.text2 = "Arcade"
        self.title = "{} {}".format(self.text1, self.text2)

    def activate(self, menu):
        pass

    def update_size(self, text):
        tw, _ = BitmapFont.title_font.measure(self.text1)
        tw2, _ = BitmapFont.title_font.measure(self.text2)
        self.w = 20 + 32 + 20 + tw + 20 + tw2 + 20

    def render_top_left(self, selected=False):
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_blending(True)
        x = 20
        y = 14
        texture = Texture.logo_32
        texture.render(x, 1080 - y - texture.h, texture.w, texture.h)
        x += 32 + 20
        BitmapFont.title_font.render(self.text1, x, self.y + 14)
        tw, _ = BitmapFont.title_font.measure(self.text1)
        x += tw + 20
        BitmapFont.title_font.render(self.text2, x, self.y + 14, alpha=0.5)
        gl.glEnable(gl.GL_DEPTH_TEST)


class HomeItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.title = gettext("Home")
        self.path_title = self.title

    def activate(self, menu):
        from arcade.glui.window import create_main_menu

        new_menu = create_main_menu()
        # State.get().history = [new_menu]
        State.get().history.append(new_menu)
        from arcade.glui.window import set_current_menu

        set_current_menu(new_menu)

    def update_size_left(self):
        self.w = 80

    def render_top_left(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        # fs_emu_blending(True)
        if selected:
            texture = Texture.home_selected
        else:
            texture = Texture.home
        texture.render(self.x, self.y, texture.w, texture.h)


class AddItem(MenuItem):
    def __init__(self):
        super().__init__()
        self.title = gettext("Add")
        self.path_title = self.title

    def update_size_left(self):
        self.w = 80

    def render_top_left(self, selected=False):
        state = State.get()
        mouse_state = state.mouse_item == self
        mouse_pressed_state = mouse_state and state.mouse_press_item == self
        self.render_top_background(
            selected,
            mouse_state=mouse_state,
            mouse_pressed_state=mouse_pressed_state,
        )
        # fs_emu_blending(True)
        if selected:
            texture = Texture.add_selected
        else:
            texture = Texture.add
        texture.render(self.x, self.y, texture.w, texture.h)

    def activate(self, menu):
        new_menu = create_item_menu(self.title)
        menu_path = self.create_menu_path(menu)
        new_menu.update_path(menu_path)
        for item in self.generate_category_items(menu_path):
            new_menu.append(item)
        return new_menu



from __future__ import absolute_import
from __future__ import print_function

from .Block import Block
from amitools.util.HexDump import *

class BitmapExtBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    # read bitmap blk ptrs
    self.bitmap_ptrs = []
    for i in xrange(self.blkdev.block_longs-1):
      bm_blk = self._get_long(i)
      self.bitmap_ptrs.append(bm_blk)
    
    self.bitmap_ext_blk = self._get_long(-1)
    
    self.valid = True
    return True
  
  def create(self):
    self.bitmap_ptrs = []
    for i in xrange(self.blkdev.block_longs-1):
      self.bitmap_ptrs.append(0)
    self.bitmap_ext_blk = 0
    self.valid = True
    return True
    
  def write(self):
    self._create_data()
    for i in xrange(self.blkdev.block_longs-1):
      self._put_long(i, self.bitmap_ptrs[i])
    self._put_long(-1, self.bitmap_ext_blk)
    self._write_data()
    
  def dump(self):
    Block.dump(self, "BitmapExtBlock", False)
    print(" bmp ptrs:  %s" % self.bitmap_ptrs)
    print(" bmp ext:   %d" % self.bitmap_ext_blk)
  


from __future__ import absolute_import
from __future__ import print_function

from .Block import Block
from .CommentBlock import CommentBlock

class EntryBlock(Block):
  """Base class for all block types that describe entries within a directory"""
  def __init__(self, blkdev, blk_num, is_type, is_sub_type, is_longname=False):
    Block.__init__(self, blkdev, blk_num, is_type, is_sub_type)
    self.is_longname = is_longname
    self.comment_block_id = 0

  def _read_nac_modts(self):
    """Reads the name, comment, and modifcation timestamp"""
    if self.is_longname:
      # In long filename mode, we have a combined field that contains
      # the filename and the comment as consequtive BSTR. If the comment does
      # not fit in, it is stored in an extra block
      nac = self._get_cstr(-46,112)
      name_len = ord(nac[0])
      self.name = nac[1:name_len+1]
      if len(nac) > name_len + 1:
        comment_len = ord(nac[name_len+1])
        self.comment = nac[name_len+2:name_len+2+comment_len]
      else:
        # Comment is located in an extra block
        self.comment_block_id = self._get_long(-18)
        self.comment = ""
      self.mod_ts = self._get_timestamp(-15)
    else:
      self.comment = self._get_bstr(-46, 79)
      self.name = self._get_bstr(-20, 30)
      self.mod_ts = self._get_timestamp(-23)

  def _write_nac_modts(self):
    """Writes the name, comment, and modifcation timestamp"""
    if self.is_longname:
      if self.comment_block_id != 0:
        nac = chr(len(self.name)) + self.name + chr(0)
      else:
        nac = chr(len(self.name)) + self.name + chr(len(self.comment)) + self.comment
      self._put_cstr(-46, 122, nac)
      self._put_long(-18, self.comment_block_id)
      self._put_timestamp(-15, self.mod_ts)
    else:
      self._put_bstr(-46, 79, self.comment)
      self._put_timestamp(-23, self.mod_ts)
      self._put_bstr(-20, 30, self.name)
  
  @staticmethod
  def needs_extra_comment_block(name, comment):
    """Returns whether the given name/comment pair requires an extra comment block"""
    return len(name) + len(comment) > 110



import platform
import sys

import fsbc.debug
import fsbc.logging

# import fsbc.unicode
from fsbc.application import Application

init_called = False
unicode_patched = False
logging_enabled = False
exception_handler_enabled = False


def initialize_application(
    name=None,
    version=None,
    patch_unicode=True,
    enable_logging=True,
    enable_exception_handler=True,
):
    global \
        init_called, \
        unicode_patched, \
        logging_enabled, \
        exception_handler_enabled
    init_called = True

    if name and enable_logging:
        fsbc.logging.setup_logging(name + ".log.txt")
        logging_enabled = True

    if name is not None:
        Application.app_name = name
        # print(name)

    if version is not None:
        Application.app_version = version
        # print(version)

    if enable_exception_handler:
        fsbc.debug.enable_exception_handler()
        exception_handler_enabled = True

    if patch_unicode:
        # patch system libraries (and argv) to work better with
        # unicode-enabled programs
        # FIXME: removed for Python 3
        # fsbc.unicode.patch_all()
        unicode_patched = True

    # print(platform.uname())
    # print(sys.argv)



from __future__ import absolute_import
from __future__ import print_function

from ..Block import Block

class LoadSegBlock(Block):
  def __init__(self, blkdev, blk_num=0):
    Block.__init__(self, blkdev, blk_num, chk_loc=2, is_type=Block.LSEG)
  
  def create(self, host_id=0, next=Block.no_blk, size=128):
    Block.create(self)
    self.size = size
    self.host_id = host_id
    self.next = next
      
  def write(self):
    if self.data == None:
      self._create_data()
    
    self._put_long(1, self.size)
    self._put_long(3, self.host_id)
    self._put_long(4, self.next)
        
    Block.write(self)
  
  def set_data(self, data):
    if self.data == None:
      self._create_data()
    self.data[20:20+len(data)] = data
    self.size = (20 + len(data)) / 4
  
  def get_data(self):
    return self.data[20:20+(self.size-5)*4]
  
  def read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    self.size = self._get_long(1)
    self.host_id = self._get_long(3)
    self.next = self._get_long(4)
    
    return self.valid
  
  def dump(self):
    Block.dump(self, "RDBlock")
    
    print(" size:           %d" % self.size)
    print(" host_id:        %d" % self.host_id)
    print(" next:           %s" % self._dump_ptr(self.next))



# Logging.py
#
# helper for logging

import logging

FORMAT='%(asctime)-15s  %(levelname)-10s  %(name)-10s  %(message)s'

def add_logging_options(parser):
  """add logging options (-v, -q, -L) to an argparse"""
  parser.add_argument('-v', '--verbose', action='count', default=0,
                      help="be more verbose")
  parser.add_argument('-q', '--quiet', action='store_true', default=False,
                      help="be totally quiet")
  parser.add_argument('-L', '--log-file', default=None,
                      help="write tool output to log file")

def setup_logging(opts):
  # setup level
  if opts.quiet:
    level = 100
  elif opts.log_file is not None:
    level = logging.DEBUG
  else:
    v = opts.verbose
    if v == 0:
      level = logging.WARNING
    elif v == 1:
      level = logging.INFO
    else:
      level = logging.DEBUG
  # setup logging
  logging.basicConfig(format=FORMAT, filename=opts.log_file, level=level)



from __future__ import absolute_import
from __future__ import print_function

import sys
import time

from .Block import Block
from ..TimeStamp import *

if sys.version_info[0] == 3:
  xrange = range


class RootBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, is_type=Block.T_SHORT, is_sub_type=Block.ST_ROOT)
  
  def create(self, name, create_ts=None, disk_ts=None, mod_ts=None, extension=0, fstype=0):
    Block.create(self)
    # init fresh hash table
    self.hash_size = self.blkdev.block_longs - 56
    self.hash_table = []
    for i in xrange(self.hash_size):
      self.hash_table.append(0)
    
    # timestamps
    self.mod_ts = mod_ts
    self.disk_ts = disk_ts
    self.create_ts = create_ts
    
    # name
    self.name = name
    
    # bitmap: blank 
    self.bitmap_flag = 0xffffffff
    self.bitmap_ptrs = []
    for i in xrange(25):
      self.bitmap_ptrs.append(0)
    self.bitmap_ext_blk = 0
    
    # new stuff for DOS6 and DOS7
    self.fstype = fstype
    self.blocks_used = 0
    
    self.extension = extension
  
  def write(self):
    self._create_data()
    
    # hash table
    self._put_long(3, self.hash_size)
    for i in xrange(self.hash_size):
      self._put_long(6+i, self.hash_table[i])
    
    # bitmap
    self._put_long(-50, self.bitmap_flag)
    for i in xrange(25):
      self._put_long(-49+i, self.bitmap_ptrs[i])
    self._put_long(-24, self.bitmap_ext_blk)
    
    # timestamps
    self._put_timestamp(-23, self.mod_ts)
    self._put_timestamp(-10, self.disk_ts)
    self._put_timestamp(-7, self.create_ts)
    
    # name
    self._put_bstr(-20, 30, self.name)
    self._put_long(-2, self.extension)
    
    # DOS6 and DOS7 stuff
    self._put_long(-11, self.blocks_used)
    self._put_long(-4, self.fstype)

    Block.write(self)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    # name hash (limit to max size)
    self.hash_size = self._get_long(3)
    
    # read (limited) hash
    hs = self.hash_size
    mhs = self.blkdev.block_longs - 56
    if hs > mhs:
      hs = mhs
    self.hash_table = []
    for i in xrange(hs):
      self.hash_table.append(self._get_long(6+i))
    
    # bitmap
    self.bitmap_flag = self._get_long(-50)
    self.bitmap_ptrs = []
    for i in xrange(25):
      bm_blk = self._get_long(-49+i)
      self.bitmap_ptrs.append(bm_blk)
    self.bitmap_ext_blk = self._get_long(-24)
    
    # timestamps
    self.mod_ts = self._get_timestamp(-23)
    self.disk_ts = self._get_timestamp(-10)
    self.create_ts = self._get_timestamp(-7)
    
    # name
    self.name = self._get_bstr(-20, 30)
    self.extension = self._get_long(-2)

    # Number of used blocks (new in DOS6 and DOS7)
    self.blocks_used = self._get_long(-11)
    # filesystem type (new in DOS6 and DOS7, 0 in others)
    self.fstype = self._get_long(-4)

    # check validity
    self.valid = True
    #self.valid = (self.bitmap_flag == 0xffffffff)
    return self.valid
  
  def dump(self):
    Block.dump(self, "Root")
    print(" hash size: %d" % self.hash_size)
    print(" hash table:%s" % self.hash_table)
    print(" bmp flag:  0x%08x" % self.bitmap_flag)
    print(" bmp ptrs:  %s" % self.bitmap_ptrs)
    print(" bmp ext:   %d" % self.bitmap_ext_blk)
    print(" mod_ts:    %s" % self.mod_ts)
    print(" disk_ts:   %s" % self.disk_ts)
    print(" create_ts: %s" % self.create_ts)
    print(" disk name: %s" % self.name)
    print(" extension: %s" % self.extension)

     


import fsui
from fsgs import openretro
from fsgs.Database import Database
from fsgs.platform import PlatformHandler
from fsgs.util.gamenameutil import GameNameUtil

from ..launcher_config import LauncherConfig
from ..launcher_settings import LauncherSettings


class ConfigurationsBrowser(fsui.VerticalItemView):
    def __init__(self, parent):
        fsui.VerticalItemView.__init__(self, parent)
        self.items = []
        self.game_icon = fsui.Image("launcher:res/16x16/controller.png")
        self.config_icon = fsui.Image("launcher:res/fsuae_config_16.png")
        LauncherSettings.add_listener(self)
        self.update_search()

        self.manual_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_yellow.png"
        )
        self.auto_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_green.png"
        )
        self.blank_icon = fsui.Image("launcher:res/16x16/blank.png")
        self.missing_color = fsui.Color(0xA8, 0xA8, 0xA8)
        self.unpublished_color = fsui.Color(0xCC, 0x00, 0x00)

        self.platform_icons = {}

    def on_destroy(self):
        LauncherSettings.remove_listener(self)

    def on_select_item(self, index):
        if index is None:
            return
        # self.load_configuration(self.items[index][str("uuid")])
        self.load_configuration(self.items[index])

    def on_activate_item(self, index):
        from ..launcherapp import LauncherApp

        LauncherApp.start_game()

    def on_setting(self, key, _):
        if key in [
            "config_search",
            "game_list_uuid",
            "database_show_games",
            "database_show_adult",
            "database_show_unpublished",
        ]:
            # if key == "game_list_uuid":
            self.update_search()
            if len(self.items) > 0:
                self.select_item(None)
                self.select_item(0)
            else:
                # self.select_item(None)
                if LauncherSettings.get("parent_uuid"):
                    LauncherSettings.set("parent_uuid", "")
                    LauncherConfig.load_default_config()
        elif key == "__config_refresh":
            self.update_search()
            self.select_item(None)
            old_parent_uuid = LauncherSettings.get("parent_uuid")
            if old_parent_uuid:
                LauncherSettings.set("parent_uuid", "")
                LauncherSettings.set("parent_uuid", old_parent_uuid)
        elif key == "parent_uuid" or key == "config_path":
            if not (
                LauncherSettings.get("parent_uuid")
                or LauncherSettings.get("config_path")
            ):
                self.select_item(None)

    def set_items(self, items):
        self.items = items
        self.update()

    def get_item_count(self):
        return len(self.items)

    def get_item_text(self, index):
        item = self.items[index]
        name = item[str("name")]
        platform = item[str("platform")] or ""
        if "[" in name:
            name, extra = name.split("[", 1)
            name = name.strip()
            extra = " \u00b7 " + extra.strip(" ]")
        else:
            extra = ""
        if fsui.toolkit == "wx":
            sep = "\n"
        else:
            sep = " \u00b7 "
            name = name.replace("\n", " \u00b7 ")
        if platform == "Amiga" and not openretro:
            platform = ""
        elif platform:
            platform = sep + PlatformHandler.get_platform_name(platform)
            # if not extra:
            #     sep = ""
            # return "{0}{1}{2}{3}".format(name, sep, extra, "")
        # else:
        text = "{0}{1}{2}".format(name, extra, platform) or "Missing Name"
        return text

        # else:
        #     return "{0} \u00b7 {1}{2}".format(name, extra, platform)

    def get_item_search_text(self, index):
        # return self.items[index][3]
        # FIXME: lower-case search string?
        return self.items[index][str("sort_key")]

    def get_item_text_color(self, index):
        have = self.items[index][str("have")]
        if not have:
            return self.missing_color
        published = self.items[index]["published"]
        if not published:
            return self.unpublished_color

    def get_item_icon(self, index):
        item = self.items[index]
        platform_id = (item[str("platform")] or "").lower()
        if item[str("have")] == 1:
            return self.manual_download_icon
        elif item[str("have")] == 0:
            return self.blank_icon
        elif item[str("have")] == 2:
            return self.auto_download_icon
        elif item[str("have")] == 4:
            if platform_id not in self.platform_icons:
                try:
                    icon = fsui.Image(
                        "launcher:res/16x16/{0}.png".format(platform_id)
                    )
                except Exception:
                    icon = self.game_icon
                self.platform_icons[platform_id] = icon
            return self.platform_icons[platform_id]
        else:
            return self.config_icon

    # def on_get_item_tooltip(self, row, column):
    #     return self.items[row][1]
    #     #text = text.replace("\nAmiga \u00b7 ", "\n")

    def update_search(self):
        search = LauncherSettings.get("config_search").strip().lower()
        print("search for", search)
        words = []
        special = []
        for word in search.split(" "):
            word = word.strip()
            if not word:
                continue
            if ":" in word[1:-1]:
                special.append(word)
            else:
                words.append(word)
        terms = GameNameUtil.extract_search_terms(" ".join(words))
        terms.update(special)

        database = Database.get_instance()

        try:
            have = int(LauncherSettings.get("database_show_games"))
        except ValueError:
            # default is show all downloadable and locally available games
            have = 1
        items = database.find_games_new(
            " ".join(terms),
            have=have,
            list_uuid=LauncherSettings.get("game_list_uuid"),
        )

        self.set_items(items)

    # noinspection PyMethodMayBeStatic
    def load_configuration(self, item):
        if item[str("uuid")]:
            LauncherSettings.set("parent_uuid", item[str("uuid")])
        else:
            config_path = Database.get_instance().decode_path(
                item[str("path")]
            )
            print("load config from", config_path)
            LauncherConfig.load_file(config_path)
            LauncherSettings.set("parent_uuid", "")



from .label import Label


class HeadingLabel(Label):
    def __init__(self, parent, label):
        Label.__init__(self, parent, label)
        font = self.get_font()
        font.set_bold(True)
        self.set_font(font)



#!/usr/bin/env python3
import sys

import launcher.version
from fsgs import OPENRETRO_DEFAULT_DATABASES, openretro
from launcher.option import Option

try:
    import typing
except ImportError:
    # Workaround to make import typing work without having it on the
    # default python path (would confuse mypy).
    import fstd.typing

    sys.modules["typing"] = fstd.typing


def find_app(app):
    if app in ["launcher", "fs-uae-launcher"]:
        from launcher.apps.fs_uae_launcher import app_main
    elif app in ["arcade", "fs-uae-arcade"]:
        from launcher.apps.fs_uae_arcade import app_main
    elif app in ["workspace"]:
        from launcher.apps.workspace import app_main
    elif app == "fs-uae-netplay-server":
        from launcher.apps.fs_uae_netplay_server import app_main
    elif app == "fs-game-center":
        from launcher.apps.fs_game_center import app_main
    elif app in ["dump-game-database", "game-database-dumper"]:
        from fsgs.gamedb.game_database_dumper import game_database_dumper_main

        app_main = game_database_dumper_main
    elif app in ["fsgs", "fs-game-runner"]:
        from launcher.apps.fsgs import app_main
    elif app == "list-plugins":
        from launcher.apps.listplugins import app_main
    elif app == "list-dirs":
        from launcher.apps.listdirs import app_main

    elif app in ["dosbox", "dosbox-fs"]:
        from launcher.apps.dosbox_fs import app_main
    elif app in ["mame", "mame-fs"]:
        from launcher.apps.mame_fs import app_main
    elif app in ["mednafen", "mednafen-fs"]:
        from launcher.apps.mednafen_fs import app_main
    elif app in ["libretro-nestopia"]:
        from launcher.apps.libretro_nestopia import app_main

    elif app in ["uade", "uade-fs"]:
        from launcher.apps.uade_fs import app_main
    elif app in ["x64sc", "x64sc-fs"]:
        from launcher.apps.x64sc_fs import app_main
    else:
        return None
    return app_main


def main():
    app_name = ""
    # Check deprecated/legacy app options.
    if "--server" in sys.argv:
        sys.argv.remove("--server")
        app_name = "fs-uae-netplay-server"
    if "--arcade" in sys.argv:
        sys.argv.remove("--arcade")
        app_name = "fs-uae-arcade"
    if "--fs-uae-arcade" in sys.argv:
        sys.argv.remove("--fs-uae-arcade")
        app_name = "fs-uae-arcade"
    if sys.argv[0].endswith("fs-game-center"):
        app_name = "fs-game-center"
    if len(sys.argv) > 1:
        if sys.argv[1] == "xdftool":
            app_name = "xdftool"
            del sys.argv[1]
    if "--xdftool" in sys.argv:
        sys.argv.remove("--xdftool")
        app_name = "xdftool"
    # Check new app option.
    for arg in sys.argv:
        if arg.startswith("--app="):
            app_name = arg[6:]
            sys.argv.remove(arg)

    import fsgs

    if "--openretro" in sys.argv:
        sys.argv.remove("--openretro")
        fsgs.product = "OpenRetro"
        fsgs.openretro = True
        for option_name in OPENRETRO_DEFAULT_DATABASES:
            Option.get(option_name)["default"] = "1"

    # Check for (fake) version override
    for arg in sys.argv:
        if arg.startswith("--") and "=" in arg:
            key, value = arg[2:].split("=", 1)
            key = key.replace("-", "_")
            if key == "fake_version":
                launcher.version.VERSION = value

    if app_name == "xdftool":
        sys.argv[0] = "xdftool"
        import amitools.tols.xdftool

        sys.exit(amitools.tools.xdftool.main())

    app_main = None
    if app_main is None:
        if app_name:
            app_main = find_app(app_name)
        elif len(sys.argv) > 1:
            app_main = find_app(sys.argv[1])
            if app_main is not None:
                # Remove app name from sys.argv
                del sys.argv[1]

    if app_main is None and not app_name:
        app_name = "fs-uae-launcher"
        app_main = find_app(app_name)
    # if openretro:
    #     if app_name == "fs-uae-launcher":
    #         app_name = "openretro-launcher"
    #     elif app_name == "fs-uae-arcade":
    #         app_name = "openretro-arcade"

    import socket

    socket.setdefaulttimeout(30.0)
    from fsbc.init import initialize_application

    initialize_application(app_name, version=launcher.version.VERSION)

    if app_main is None:
        print("No valid app specified", file=sys.stderr)
        sys.exit(1)
    app_main()



from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from arcade.arcadetheme import ArcadeTheme
from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from OpenGL import GL as gl

# import numpy
# from PIL import Image
# from fsbc.Application import app
from arcade.resources import resources
from fsui.qt import QImage


class Texture(object):
    shadow = None
    shadow2 = None
    gloss = None
    screen_gloss = None
    static = None
    default_item = None
    missing_screenshot = None
    missing_cover = None
    logo = None
    top = None
    top_logo = None
    top_logo_selected = None
    splash = None

    add = None
    add_selected = None
    home = None
    home_selected = None
    minimize = None
    minimize_selected = None
    close = None
    close_selected = None
    shutdown = None
    shutdown_selected = None

    bottom_bar = None
    screen_border_1 = None
    screen_border_2 = None
    top_background = None
    top_item = None
    top_item_selected = None
    top_item_left = None
    top_item_left_selected = None
    top_item_right = None
    top_item_arrow = None
    top_item_arrow_selected = None

    glow_top = None
    glow_top_left = None
    glow_left = None

    sidebar_background = None
    sidebar_background_shadow = None
    heading_strip = None
    item_background = None
    top_item_background = None
    logo_32 = None

    stretch = None
    aspect = None
    square_pixels = None

    def __init__(self, name, target=gl.GL_TEXTURE_2D, **kwargs):
        # print(repr(type(name)))
        # if isinstance(name, (int, long)):
        if not isinstance(name, str):
            self.size = kwargs["size"]
            self.texture = name
        else:
            self.size = [0, 0]
            # print(name, kwargs)
            out_data = {}
            self.texture = self.from_theme(
                name,
                target=target,
                size=self.size,
                out_data=out_data,
                **kwargs,
            )
            self.data = out_data["im_data"]
            self.gl_type = out_data["type"]
        self.w, self.h = self.size
        self.target = target

    def bind(self):
        gl.glBindTexture(self.target, self.texture)

    def render(self, x, y, w=None, h=None, z=0.0, opacity=1.0):
        if w is None:
            w = self.w
        if h is None:
            h = self.h
        self.bind()
        fs_emu_texturing(True)
        if self.gl_type == gl.GL_RGBA or opacity < 1.0:
            fs_emu_blending(True)
        else:
            fs_emu_blending(False)
        gl.glBegin(gl.GL_QUADS)
        if opacity < 1.0:
            gl.glColor4f(opacity, opacity, opacity, opacity)
        else:
            gl.glColor3f(1.0, 1.0, 1.0)
        gl.glTexCoord(0.0, 1.0)
        gl.glVertex3f(x, y, z)
        gl.glTexCoord(1.0, 1.0)
        gl.glVertex3f(x + w, y, z)
        gl.glTexCoord(1.0, 0.0)
        gl.glVertex3f(x + w, y + h, z)
        gl.glTexCoord(0.0, 0.0)
        gl.glVertex3f(x, y + h, z)
        gl.glEnd()

    @classmethod
    def load(
        cls,
        im,
        mipmap=False,
        min_filter=None,
        wrap_s=gl.GL_CLAMP_TO_EDGE,
        wrap_t=gl.GL_CLAMP_TO_EDGE,
        target=gl.GL_TEXTURE_2D,
        size=None,
        out_data=None,
    ):
        if size is None:
            size = [0, 0]
        # type = "RGB"
        # gl_type = gl.GL_RGB
        # if im.mode == "RGBA":
        #     # convert to premultiplied alpha
        #     #pix = im.load()
        #     #for x in range(im.size[0]):
        #     #    for y in range(im.size[1]):
        #     #        r, g, b, a = pix[x, y]
        #     #        if a:
        #     #            pix[x, y] = r * 255 // a, g * 255 // a, \
        #     #                 b * 255 // a, a
        #     #        else:
        #     #            pix[x, y] = 0, 0, 0, 0
        #     a = numpy.fromstring(im.tostring(), dtype=numpy.uint8)
        #     alpha_layer = a[3::4] / 255.0
        #     a[0::4] *= alpha_layer
        #     a[1::4] *= alpha_layer
        #     a[2::4] *= alpha_layer
        #     #im = Image.fromstring("RGBA", im.size, a.tostring())
        #     im_data = a.tostring()
        #     # type = "RGBA"
        #     gl_type = gl.GL_RGBA
        # else:
        #     im_data = im.tostring("raw", "RGB")
        #
        # size[0] = im.size[0]
        # size[1] = im.size[1]
        # #texture = glGenTextures(1)

        internal_format = gl.GL_RGBA
        texture_format = gl.GL_BGRA

        if im.format() != QImage.Format.Format_ARGB32_Premultiplied:
            im = im.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)

        bits = im.bits()
        try:
            pixels = bits.tobytes()
        except AttributeError:
            bits.setsize(im.sizeInBytes())
            pixels = bytes(bits)
        size[0] = im.width()
        size[1] = im.height()

        from arcade.glui.render import Render

        texture = Render.get().create_texture()
        gl.glBindTexture(target, texture)
        gl.glTexImage2D(
            target,
            0,
            internal_format,
            size[0],
            size[1],
            0,
            texture_format,
            gl.GL_UNSIGNED_BYTE,
            pixels,
        )
        if mipmap:
            gl.glGenerateMipmap(target)
            if min_filter is None:
                min_filter = gl.GL_LINEAR_MIPMAP_LINEAR
        else:
            if min_filter is None:
                min_filter = gl.GL_LINEAR
        gl.glTexParameteri(target, gl.GL_TEXTURE_MIN_FILTER, min_filter)
        gl.glTexParameteri(target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        gl.glTexParameteri(target, gl.GL_TEXTURE_WRAP_S, wrap_s)
        gl.glTexParameteri(target, gl.GL_TEXTURE_WRAP_T, wrap_t)
        if out_data is not None:
            out_data["im_data"] = pixels
            out_data["type"] = internal_format
        return texture

    @classmethod
    def from_resource(cls, name, size=None, **kwargs):
        if size is None:
            size = [0, 0]
        im = resources.resource_qt_image(name)
        return cls.load(im, size=size, **kwargs)

    # @classmethod
    # def from_qimage(cls, name, size=None, **kwargs):
    #     if size is None:
    #         size = [0, 0]
    #     im = resources.resource_qt_image(name)
    #     return cls.load(im, size=size, **kwargs)

    @classmethod
    def from_theme(cls, name, size=None, **kwargs):
        if size is None:
            size = [0, 0]
        # im = resources.resource_qt_image(name)
        im = ArcadeTheme.get().qimage(name)
        return cls.load(im, size=size, **kwargs)



import os

import fsui
from fsgs.FSGSDirectories import FSGSDirectories

from ..launcher_settings import LauncherSettings


class LauncherFilePicker(object):
    def __init__(
        self,
        parent,
        title,
        media_type,
        last_path="",
        multiple=False,
        dir_mode=False,
    ):
        self.multiple = multiple
        self.dir_mode = dir_mode
        self.parent = parent
        self.title = title
        self.result = None
        self.dir_mode = dir_mode
        self.settings_key = "last_{0}_dir".format(media_type)
        self.directory = ""
        if last_path and last_path not in ["internal"]:
            print("last_path", repr(last_path))
            # if os.path.isdir(last_path):
            #     last_path_dir = last_path
            # else:
            last_path_dir = os.path.dirname(last_path)
            print("last_path_dir", last_path_dir)
            if last_path_dir:
                if os.path.exists(last_path_dir):
                    self.directory = last_path_dir
            else:
                # file was relative to default directory
                self.directory = self.get_default_directory(media_type)
        if not self.directory:
            value = LauncherSettings.get(self.settings_key)
            print(self.settings_key, value)
            if value and os.path.exists(value):
                self.directory = value
        if not self.directory:
            self.directory = self.get_default_directory(media_type)
        # fsui.FileDialog.__init__(
        #     self, parent, title, directory, dir_mode=dir_mode,
        #     multiple=multiple)

    @staticmethod
    def get_default_directory(media_type):
        if media_type == "floppy":
            return FSGSDirectories.get_floppies_dir()
        elif media_type == "cd":
            return FSGSDirectories.get_cdroms_dir()
        elif media_type == "hd":
            return FSGSDirectories.get_hard_drives_dir()
        elif media_type == "rom":
            return FSGSDirectories.get_kickstarts_dir()
        # FIXME: Return new Media directory instead
        return FSGSDirectories.get_base_dir()
        # raise Exception("unknown file dialog type")

    def show_modal(self):
        path = None
        if self.dir_mode:
            self.result = fsui.pick_directory(
                parent=self.parent,
                message=self.title,
                directory=self.directory,
            )
            if self.result:
                path = self.result
        elif self.multiple:
            self.result = fsui.pick_files(
                parent=self.parent,
                message=self.title,
                directory=self.directory,
            )
            if self.result:
                path = self.result[0]
        else:
            self.result = fsui.pick_file(
                parent=self.parent,
                message=self.title,
                directory=self.directory,
            )
            if self.result:
                path = self.result
        if path:
            last_path_dir = os.path.dirname(path)
            LauncherSettings.set(self.settings_key, last_path_dir)
        return bool(self.result)

    def destroy(self):
        pass

    def get_path(self):
        return self.result

    def get_paths(self):
        return self.result



from __future__ import absolute_import
from __future__ import print_function

import time
from .Block import Block
from .EntryBlock import EntryBlock
from .CommentBlock import CommentBlock
from ..ProtectFlags import ProtectFlags
from ..TimeStamp import *

class FileHeaderBlock(EntryBlock):
  def __init__(self, blkdev, blk_num, is_longname):
    EntryBlock.__init__(self, blkdev, blk_num, is_type=Block.T_SHORT, is_sub_type=Block.ST_FILE, is_longname=is_longname)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    # FileHeader fields
    self.own_key = self._get_long(1)
    self.block_count = self._get_long(2)
    self.first_data = self._get_long(4)
    
    # read (limited) data blocks table
    bc = self.block_count
    mbc = self.blkdev.block_longs - 56
    if bc > mbc:
      bc = mbc
    self.data_blocks = []
    for i in xrange(bc):
      self.data_blocks.append(self._get_long(-51-i))
    
    self.protect = self._get_long(-48)
    self.protect_flags = ProtectFlags(self.protect)
    self.byte_size = self._get_long(-47)
    self._read_nac_modts()
    self.hash_chain = self._get_long(-4)
    self.parent = self._get_long(-3)
    self.extension = self._get_long(-2)

    self.valid = (self.own_key == self.blk_num)
    return self.valid
  
  def write(self):
    Block._create_data(self)
    self._put_long(1, self.own_key)
    self._put_long(2, self.block_count)
    self._put_long(4, self.first_data)
    
    # data blocks
    for i in xrange(len(self.data_blocks)):
      self._put_long(-51-i, self.data_blocks[i])
    
    self._put_long(-48, self.protect)
    self._put_long(-47, self.byte_size)

    self._write_nac_modts()

    self._put_long(-4, self.hash_chain)
    self._put_long(-3, self.parent)
    self._put_long(-2, self.extension)
    Block.write(self)
  
  def create(self, parent, name, data_blocks, extension, byte_size=0, protect=0, comment=None, mod_ts=None, hash_chain=0):
    Block.create(self)
    self.own_key = self.blk_num
    n = len(data_blocks)
    self.block_count = n
    if n == 0:
      self.first_data = 0
    else:
      self.first_data = data_blocks[0]
    
    self.data_blocks = data_blocks
    self.protect = protect
    self.protect_flags = ProtectFlags(self.protect)
    self.byte_size = byte_size
    if comment == None:
      self.comment = ''
    else:
      self.comment = comment
    self.mod_ts = mod_ts
    self.name = name
    self.hash_chain = hash_chain
    self.parent = parent
    self.extension = extension
    self.valid = True
    return True
    
  def dump(self):
    Block.dump(self,"FileHeader")
    print(" own_key:    %d" % self.own_key)
    print(" blk_cnt:    %d" % self.block_count)
    print(" first_data: %d" % self.first_data)
    if self.data_blocks != None:
      print(" data blks:  %s #%d" % (self.data_blocks, len(self.data_blocks)))
    pf = ProtectFlags(self.protect)
    print(" protect:    0x%x 0b%s %s" % (self.protect, pf.bin_str(), pf))
    print(" byte_size:  %d" % self.byte_size)
    print(" comment:    '%s'" % self.comment)
    print(" mod_ts:     %s" % self.mod_ts)
    print(" name:       '%s'" % self.name)
    print(" hash_chain: %d" % self.hash_chain)
    print(" parent:     %d" % self.parent)
    print(" extension:  %d" % self.extension)
  


import os
import shutil
import traceback
import fsui
from fsgs.FSGSDirectories import FSGSDirectories

# from workspace.shell import SimpleApplication
from launcher.res import gettext
from fsui.extra.iconheader import IconHeader
from workspace.ui.theme import WorkspaceTheme


class HDFCreatorWindow(fsui.DialogWindow):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        title = gettext("HDF Disk Image Creator")
        super().__init__(parent, title=title)
        self.theme = WorkspaceTheme.instance()
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        if self.window.theme.has_close_buttons:
            buttons.create_close_button()

        self.dialog = None
        self.path = ""

        self.set_icon(fsui.Icon("hd-volume", "pkg:workspace"))

        layout.min_width = 500

        self.icon_header = IconHeader(
            self,
            fsui.Icon("hd-volume", "pkg:workspace"),
            gettext("HDF Disk Image Creator"),
            gettext(
                "Create a single-partition or partitionable hard drive image"
            ),
        )
        layout.add(self.icon_header, fill=True, margin_bottom=20)

        label = fsui.Label(self, gettext("Create disk image of type:"))
        layout.add(label)
        layout.add_spacer(6)
        self.list_view = fsui.ListView(self)
        self.list_view.set_min_width(560)
        self.list_view.set_min_height(60)
        icon = fsui.Image("workspace:res/16x16/hd-volume.png")
        self.list_view.add_item(
            gettext("HDF - Single Partition Hard Disk File"), icon
        )
        self.list_view.add_item(
            gettext("HDF - Partitionable Hard Drive Image (RDB)"), icon
        )
        layout.add(self.list_view, expand=True, fill=True)
        self.list_view.item_selected.connect(self.on_item_selected)

        layout.add_spacer(20)
        label = fsui.Label(self, gettext("Filename for the new disk image:"))
        layout.add(label)
        layout.add_spacer(6)
        hori_layout = fsui.HorizontalLayout()
        layout.add(hori_layout, fill=True)
        self.name_field = fsui.TextField(self, "", read_only=False)
        hori_layout.add(self.name_field, expand=True)
        text = gettext("Size:")
        label = fsui.Label(self, text)
        hori_layout.add(label, margin_left=20)
        self.size_field = fsui.TextField(self, "")
        self.size_field.set_min_width(60)
        hori_layout.add(self.size_field, expand=False, margin_left=10)
        text = gettext("MB")
        label = fsui.Label(self, text)
        hori_layout.add(label, margin_left=10)

        layout.add_spacer(20)
        label = fsui.Label(self, gettext("Save to directory:"))
        layout.add(label)
        layout.add_spacer(6)
        hori_layout = fsui.HorizontalLayout()
        layout.add(hori_layout, fill=True)
        self.dir_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.dir_field, expand=True)
        self.browse_button = fsui.Button(self, gettext("Browse"))
        self.browse_button.clicked.connect(self.on_browse_clicked)
        hori_layout.add(self.browse_button, margin_left=10)

        self.created_label = fsui.Label(self, "")
        layout.add(self.created_label, fill=True)

        # layout.add_spacer(20)
        # layout.add_spacer(20)
        # hori_layout = fsui.HorizontalLayout()
        # layout.add(hori_layout, fill=True)
        # self.created_label = fsui.Label(self, "")
        # hori_layout.add(self.created_label, expand=True)
        # hori_layout.add_spacer(20)
        self.create_button = fsui.Button(buttons, gettext("Create"))
        # self.create_button.activated.connect(self.on_create_clicked)
        self.create_button.clicked.connect(self.on_create_clicked)
        # hori_layout.add(self.create_button)
        buttons.add_button(self.create_button)

        self.list_view.select_item(0)
        self.update_name_suggestion()

        # self.set_size(self.layout.get_min_size())
        # self.center_on_parent()

    def __del__(self):
        print("HDFCreator.__del__")

    def update_name_suggestion(self):
        k = 1
        while True:
            if k > 1:
                name = "New Disk {0}.hdf".format(k)
            else:
                name = "New Disk.hdf"
            if not os.path.exists(os.path.join(self.path, name)):
                self.name_field.set_text(name)
                return
            k += 1

    def on_item_selected(self, index):
        self.set_path(FSGSDirectories.get_hard_drives_dir())
        if index == 0:
            size = "256"
        elif index == 1:
            size = "1024"
        else:
            raise Exception("Unexpected index " + repr(index))
        self.size_field.set_text(size)
        self.size_field.enable(True)

    def set_path(self, path):
        self.path = path
        self.dir_field.set_text(path)

    def on_browse_clicked(self):
        # if hasattr(self, "dialog"):
        #    return
        self.dialog = fsui.DirDialog(
            None, gettext("Select Destination Directory")
        )
        self.dialog.accepted.connect(self.on_dialog_accepted)
        # self.dialog.destroyed.connect(self.on_dialog_destroyed)
        # self.dialog.finished.connect(self.on_dialog_finished)
        self.dialog.show()
        # if dialog.show_modal():
        #     self.set_path(dialog.get_path())
        # dialog.destroy()

    def on_dialog_accepted(self):
        print("accepted")
        print(self.dialog)
        self.set_path(self.dialog.get_path())
        self.dialog = None

    # def on_dialog_destroyed(self):
    #     self.dialog = None

    # def on_dialog_finished(self, result):
    #     # del self.dialog
    #     # self.dialog = None
    #     # return
    #     print("finished")
    #     if self.dialog.result():
    #         print("result")
    #         self.set_path(self.dialog.get_path())
    #     # self.dialog.destroy()
    #     # print("...")
    #     del self.dialog

    def on_create_clicked(self):
        try:
            self.create_disk_file()
        except Exception as e:
            traceback.print_exc()
            self.show_error(repr(e))

    def show_error(self, message):
        self.created_label.set_text(message)
        fsui.show_error(message)

    def show_success(self, message):
        self.created_label.set_text(message)

    def create_disk_file(self):
        self.created_label.set_text("")
        disk_type = self.list_view.get_index()
        path = self.dir_field.get_text().strip()
        if not os.path.isdir(path):
            return self.show_error(
                gettext("Specified directory does not exist")
            )
        name = self.name_field.get_text().strip()
        ext = ".hdf"
        if not name.lower().endswith(ext):
            name += ext
        path = os.path.join(path, name)

        try:
            size = int(self.size_field.get_text())
        except ValueError:
            return self.show_error(gettext("Invalid size specified"))
        else:
            size = size * 1024 * 1024
        if disk_type == 2:
            if size >= 512 * 1024 * 1024:
                return self.show_error(
                    gettext("Use RDB disk images for size >= {0}".format(512))
                )

        if os.path.exists(path):
            return self.show_error(gettext("File already exists"))

        path_partial = path + ".partial"
        try:
            f = open(path_partial, "wb")
        except Exception:
            print("error opening", path)
            traceback.print_exc()
            return self.show_error(gettext("Could not open file for writing"))

        try:
            if disk_type == 0:
                self.create_hdf(f, size)
            elif disk_type == 1:
                self.create_rdb(f, size)
        except Exception:
            traceback.print_exc()
            try:
                f.close()
            except Exception:
                pass
            try:
                os.unlink(path_partial)
            except Exception:
                pass
            return self.show_error(gettext("Error writing disk image"))
        else:
            f.close()
            try:
                shutil.move(path_partial, path)
            except Exception:
                print("error moving", path_partial, path)
                traceback.print_exc()
                return self.show_error(gettext("Error moving file into place"))
        self.show_success(gettext("Disk image created") + ": " + name)

    def create_hdf(self, f, size):
        print("create_hdf", f, size)
        assert size % (1024 * 1024) == 0
        block = b"\0" * 1024 * 1024
        for i in range(size // (1024 * 1024)):
            f.write(block)

    def create_rdb(self, f, size):
        print("create_rdb", f, size)
        assert size % (1024 * 1024) == 0
        block = b"\0" * 1024 * 1024
        for i in range(size // (1024 * 1024)):
            f.write(block)
        f.seek(0)
        f.write(b"rdsk")


# application = SimpleApplication(HDFCreatorWindow)
